from pprint import pformat
import warnings
from typing import Optional

from contracts import contract
from rosbag_utils import resolve_topics, rosbag_info_cached

from procgraph import BadConfig, Block, Generator

__all__ = ["BagRead"]


class BagRead(Generator):
    """
        This block reads a bag file (ROS logging format).
    """

    Block.alias("bagread")
    Block.output_is_defined_at_runtime("The signals read from the log.")
    Block.config("file", "Bag file to read")
    Block.config("limit", "Limit in seconds on how much data we want. (0=no limit)", default=0)
    Block.config("t0", "Relative start time", default=None)
    Block.config("t1", "Relative start time", default=None)

    Block.config(
        "topics",
        "Which signals to output (and in what order). "
        "Should be a comma-separated list. If you do not specify it "
        "(or if empty) it will be all signals.",
        default=[],
    )

    Block.config("quiet", "If true, disables advancements status messages.", default=False)

    @contract(returns="list(str)")
    def get_topics(self):
        bagfile = self.config.file
        if self.config.topics is not None:
            given_topics = self.config.topics.strip()
        else:
            given_topics = None

        if given_topics:
            topics = given_topics.split(",")
        else:
            all_topics = [c.topic for c in self.bag._get_connections()]
            topics = sorted(set(all_topics))

        self.baginfo = rosbag_info_cached(bagfile)
        res, _, asked2resolved = resolve_topics(self.baginfo, topics)
        self.info("Resolving:\n%s" % pformat(asked2resolved))
        return res

    def get_output_signals(self):
        import rosbag

        self.bag = rosbag.Bag(self.config.file)

        self.topics = self.get_topics()

        self.topic2signal = {}
        signals = []
        for t in self.topics:
            self.info(t)
            if ":" in t:
                tokens = t.split(":")
                assert len(tokens) == 2
                t = tokens[0]
                signal_name = tokens[1]
            else:
                signal_name = str(t).split("/")[-1]
            if signal_name in signals:
                self.error("Warning: repeated name %s" % signal_name)
                signal_name = ("%s" % t).replace("/", "_")
                self.error("Using long form %r." % signal_name)
            signals.append(signal_name)
            self.topic2signal[t] = signal_name

        topics = list(self.topic2signal.keys())

        self.info(self.topic2signal)

        limit = self.config.limit
        if not isinstance(limit, (float, int)):
            msg = "I require a number; 0 for none."
            raise BadConfig(msg, self, "limit")

        if self.config.t0 is not None or self.config.t1 is not None:
            t0 = self.config.t0
            t1 = self.config.t1
            start_time, end_time = self._get_start_end_time_t0_t1(t0, t1)
        else:
            start_time, end_time = self._get_start_end_time(limit)

        print(("t0: %s" % self.config.t0))
        print(("t1: %s" % self.config.t1))
        print(("start_time: %s" % start_time))
        print(("end_time: %s" % end_time))
        print(("start_stamp: %s" % self.start_stamp))
        print(("end_stamp: %s" % self.end_stamp))

        params = dict(topics=topics, start_time=start_time, end_time=end_time)

        self.iterator = self.bag.read_messages(**params)

        return signals

    def _get_start_end_time(self, limit: Optional[float]):
        """
            Returns the start and end time to use (rospy.Time).

            also sets self.start_stamp, self.end_stamp

        """
        from rospy.rostime import Time

        self.info("limit: %r" % limit)
        warnings.warn("Make better code for dealing with unindexed")
        if limit is not None and limit != 0:
            # try:
            chunks = self.bag.__dict__["_chunks"]
            self.start_stamp = chunks[0].start_time.to_sec()
            self.end_stamp = chunks[-1].end_time.to_sec()
            start_time = Time.from_sec(self.start_stamp)
            end_time = Time.from_sec(self.start_stamp + limit)
            return start_time, end_time
            # except Exception as e:
            #     self.error('Perhaps unindexed bag?')
            #     self.error(traceback.format_exc(e))
            #     raise
            #     start_time = None
            #     end_time = None
            #
            # self.info('start_stamp: %s' % self.start_stamp)
            # self.info('end_stamp: %s' % self.end_stamp)
        else:
            self.start_stamp = None
            self.end_stamp = None
            return None, None

    def _get_start_end_time_t0_t1(self, t0, t1):
        """
            Returns the start and end time to use (rospy.Time).

            also sets self.start_stamp, self.end_stamp

        """
        warnings.warn("Make better code for dealing with unindexed")
        from rospy.rostime import Time

        if t0 is not None or t1 is not None:
            # try:
            chunks = self.bag.__dict__["_chunks"]
            self.start_stamp = chunks[0].start_time.to_sec()
            self.end_stamp = chunks[-1].end_time.to_sec()
            if t0 is not None:
                start_time = self.start_stamp + t0
            else:
                start_time = self.start_stamp
            if t1 is not None:
                end_time = self.start_stamp + t1
            else:
                end_time = self.end_stamp
            start_time = Time.from_sec(start_time)
            end_time = Time.from_sec(end_time)
            return start_time, end_time
            # except Exception as e:
            #     self.error('Perhaps unindexed bag?')
            #     self.error(traceback.format_exc(e))
            #     raise
            #     start_time = None
            #     end_time = None
            #
            # self.info('start_stamp: %s' % self.start_stamp)
            # self.info('end_stamp: %s' % self.end_stamp)
        else:
            self.start_stamp = None
            self.end_stamp = None
            return None, None

    def init(self):
        self._load_next()

    def _load_next(self):
        try:
            topic, msg, timestamp = next(self.iterator)
            self.next_timestamp = timestamp.to_sec()
            self.next_value = msg
            self.next_topic = topic
            self.next_signal = self.topic2signal[topic]
            self.has_next = True
        except StopIteration:
            self.has_next = False

    def next_data_status(self):
        if self.has_next:
            return (self.next_signal, self.next_timestamp)
        else:
            return (False, None)

    def update(self):
        if not self.has_next:
            return  # XXX: error here?

        self.set_output(self.next_signal, value=self.next_value, timestamp=self.next_timestamp)

        self._load_next()

    def finish(self):
        self.bag.close()



import numpy as np
from numpy import multiply, array

from procgraph import COMPULSORY, register_simple_block, simple_block
from procgraph.core.block import Block


@simple_block
def astype(a, dtype=COMPULSORY):
    """
        Converts an array using the ``astype`` function.

        :param a: Numpy array
        :type a: array

        :param dtype: The new dtype.
        :type dtype: string

        :return: typed: The Numpy array with the new type.
        :rtype: array
    """
    return a.astype(dtype)


@simple_block
def take(a, axis=0, indices=COMPULSORY):
    assert indices != COMPULSORY
    a = np.array(a)
    indices = list(indices)  # parsingresult bug
    axis = int(axis)
    try:
        return a.take(axis=axis, indices=indices).squeeze()
    except Exception as e:
        raise Exception(
            "take(axis=%s,indices=%s) failed on array " "with shape %s: %s" % (axis, indices, a.shape, e)
        )


@simple_block
def sub(a, f=COMPULSORY):
    """ Takes a field """
    return a[f]


@simple_block
def outer(a, b):
    """
        Outer product of two vectors.

        This is a wrapper around :py:func:`np.multiply.outer`.

        :param a: First vector.
        :param b: Second vector.
        :return: outer: Outer product of the two vectors.
    """
    a = array(a)
    b = array(b)
    res = multiply.outer(a, b)
    return res


@simple_block
def select(x, every=COMPULSORY):
    """
        Selects some of the elements of ``x``.

        :param x: Numpy array that can be flatly addressed.
        :param every: How many to jump (every=2 takes only the even elements).
        :return: decimated: The decimated output.
    """
    assert every != COMPULSORY
    n = len(x)
    return x[list(range(0, n, every))]


@simple_block
def normalize_Linf(x):
    """ Normalize a vector such that ``|x|_inf = max(abs(x))= 1``.

        :param x: Any numpy array.
        :return: normalized: The same array normalized.

    """
    return x / np.abs(x).max()


@simple_block
def minimum(value, threshold=COMPULSORY):
    """ Limits the numpy array to the given threshold.

        :param value: Any numpy array.
        :return: Array of same shape.
    """
    assert threshold != COMPULSORY
    return np.minimum(value, threshold)


@simple_block
def maximum(value, threshold=COMPULSORY):
    """ Limits the numpy array to the given threshold. """
    assert threshold != COMPULSORY
    return np.maximum(value, threshold)


@simple_block
def norm(value, ord=None):  # @ReservedAssignment
    """ Returns the norm of the vector.

        ord=None  Frobenius for matrix    2-norm for vectors
        2         2-norm (largest sing.value)
    """
    return np.linalg.norm(value, ord)


@simple_block
def norm2(value):  # @ReservedAssignment
    """ Returns the norm of the vector. """
    return np.linalg.norm(value)


# XXX: not sure the best thing to do
@simple_block
def array(value):
    """ Converts the value to a Numpy array. """
    return np.array(value)


@simple_block
def list2array(value):
    """ Converts a list of uniform elements to a Numpy array. """
    return np.array(value, dtype=value[0].dtype)


register_simple_block(np.mean, "mean", params={"axis": 0}, doc="Wrapper around :np:data:`np.mean`.")

register_simple_block(np.square, "square", doc="Wrapper around :np:data:`np.square`.")

register_simple_block(np.log, "log", doc="Wrapper around :np:data:`np.log`.")

register_simple_block(np.abs, "abs", doc="Wrapper around :np:data:`np.absolute`.")

register_simple_block(np.sign, "sign", doc="Wrapper around :np:data:`np.sign`.")

register_simple_block(np.arctan, "arctan", doc="Wrapper around :np:data:`np.arctan`.")

register_simple_block(np.real, "real", doc="Wrapper around :np:data:`np.real`.")


register_simple_block(
    lambda x, y: np.dstack((x, y)), "dstack", num_inputs=2, doc="Wrapper around :np:data:`np.ma.dstack`."
)

register_simple_block(
    lambda x, y: np.hstack((x, y)), "hstack", num_inputs=2, doc="Wrapper around :np:data:`np.ma.hstack`."
)

register_simple_block(
    lambda x, y: np.vstack((x, y)), "vstack", num_inputs=2, doc="Wrapper around :np:data:`np.ma.vstack`."
)

register_simple_block(lambda x: np.max(array(x).flat), "max", doc="Maximum over **all** elements. ")

register_simple_block(lambda x: np.sum(array(x).flat), "sum", doc="Sum over **all** elements. ")

register_simple_block(np.sum, "sum_axis", doc="Sum over the axis. ")


register_simple_block(np.flipud, "flipud", doc="Flips the array up/down (wrapper for :py:func:`np.flipud`.)")

register_simple_block(
    np.fliplr, "fliplr", doc="Flips the array left/right (wrapper for :py:func:`np.fliplr`.)"
)

register_simple_block(np.transpose, "transpose", doc="Tranpose (wrapper for :py:func:`np.transpose`.)")


register_simple_block(
    np.radians, "deg2rad", doc="Converts degrees to radians (wrapper around " ":np:data:`np.radians`.)"
)

register_simple_block(
    np.degrees, "rad2deg", doc="Converts radians to degrees (wrapper around " ":np:data:`np.degrees`.)"
)


class HSplit(Block):
    """ Splits an array along the first axis. """

    Block.alias("hsplit")

    Block.input("value")

    Block.output("half1")
    Block.output("half2")

    def update(self):
        value = self.input.value
        h = int(value.shape[0] / 2)
        self.output.half1 = value[0:h, ...]
        self.output.half2 = value[h:, ...]



from ..core.constants import PATH_ENV_VAR
from ..core.exceptions import BadMethodCall, PGException, SemanticError
from ..core.model import Model
from ..core.model_loader import ModelSpec, model_from_string, pg_look_for_models
from ..core.registrar import Library, default_library
from ..core.visualization import error, info
from contracts import contract
from optparse import OptionParser
import os
import sys
import traceback
import yaml


usage_short = """Usage:

    pg [options]  <model>.pg   [param=value  param=value ... ]

Type "pg --help" for all the options and a few examples.

"""

usage_long = (
    usage_short
    + """Examples:

1) Execute a model that does not need parameters:

    $ pg  my_model.pg

2) Execute a model, reading the a directory for additional models:

    $ pg -d my_models/  my_model

   (Note that the current directory is not read by default).
   There is also an environment variable that has the same effect:

    $ export {PATH_ENV_VAR}=my_models

3) Execute a model, but first load a module that might contain additional block
   definitions.

    $ pg -m my_blocks  my_model.pg      """.format(
        PATH_ENV_VAR=PATH_ENV_VAR
    )
)


def main():

    # TODO: use LenientOptionParser
    parser = OptionParser(usage=usage_long)
    parser.disable_interspersed_args()
    additional_modules = []

    def load_module(option, opt_str, value, parser):  # @UnusedVariable
        additional_modules.append(value)

    additional_directories = []

    def add_directory(option, opt_str, value, parser):  # @UnusedVariable
        additional_directories.append(value)

    parser.add_option(
        "-m",
        dest="module",
        action="callback",
        callback=load_module,
        type="string",
        help="Loads the specified module.",
    )

    parser.add_option(
        "-d",
        dest="directory",
        type="string",
        action="callback",
        callback=add_directory,
        help="Additional directory to search for models.",
    )

    parser.add_option(
        "--debug", action="store_true", default=False, help="Displays debug information on the model."
    )

    parser.add_option(
        "--trace",
        action="store_true",
        default=False,
        dest="trace",
        help="If true, try to display raw stack trace in case  "
        " of error, instead of the usual friendly message.",
    )

    parser.add_option(
        "--stats",
        action="store_true",
        default=False,
        dest="stats",
        help="Displays execution stats, including CPU usage.",
    )

    parser.add_option("--nocache", action="store_true", default=False, help="Ignores the parsing cache.")

    (options, args) = parser.parse_args()

    # TODO: make an option to display all the known models
    # if options.debug:
    #    print "Configuration: %s" % config
    if not args:
        print(usage_short)
        sys.exit(-1)

    filename = args.pop(0)

    if options.trace:
        look_for = RuntimeError
    else:
        look_for = PGException

    try:
        config = parse_cmdline_args(args)

        pg(
            filename,
            config,
            nocache=options.nocache,
            debug=options.debug,
            stats=options.stats,
            additional_directories=additional_directories,
            additional_modules=additional_modules,
        )

        sys.exit(0)

    except look_for as e:
        error(e)
        if not options.trace:
            info(
                'If you run "pg" with the "--trace" option, you can see the ' "python details for this error."
            )
        sys.exit(-2)


def parse_cmdline_args(args):
    """ Parses the command-line arguments into a configuration dictionary. """
    config = {}

    def found_pair(key, value_string):
        try:
            value = int(value_string)
        except:
            try:
                value = float(value_string)
            except:
                value = value_string

        # value = parse_value(value_string)
        config[key] = value

    while args:
        arg = args.pop(0)
        if "=" in arg:
            key, value_string = arg.split("=")
            found_pair(key, value_string)
        elif arg.startswith("--"):
            key = arg[2:]
            if not args:
                msg = "Argument for %r missing." % key
                raise Exception(msg)
            value_string = args.pop(0)
            found_pair(key, value_string)
        else:
            msg = "I don't know how to interpret %r" % arg
            raise Exception(msg)

    return config


@contract(
    filename="str",
    stats="bool",
    nocache="bool",
    stats_display_interval="None|int,>0",
    additional_directories="list(str)",
    additional_modules="list(str)",
    config="dict(str:*)",
    returns=Model,
)
def pg(
    filename,
    config,
    debug=False,
    nocache=False,
    stats=False,
    stats_display_interval=100,
    additional_directories=[],
    additional_modules=[],
):
    """
        Instantiate and run a model.

        :param filename: either a file or a known model

        :param stats_display_interval: how many iterations to display stats.
    """

    for module in additional_modules:
        info("Importing package %r..." % module)
        __import__(module)

    library = Library(default_library)

    pg_look_for_models(library, ignore_cache=nocache, additional_paths=additional_directories)

    # load standard components
    # noinspection PyUnresolvedReferences
    import procgraph.components  # @UnusedImport

    if library.exists(block_type=filename):
        # XXX w = Where('command line', filename, 0)
        # Check that it is a model, and not a block.
        generator = library.get_generator_for_block_type(filename)
        if not isinstance(generator, ModelSpec):
            # XXX nothing given
            msg = (
                "The name %r corresponds to a block, not a model. "
                'You can only use "pg" with models. ' % filename
            )
            raise SemanticError(msg, None)

        if len(generator.input) > 0:
            inputs = ", ".join([x.name.__repr__() for x in generator.input])
            msg = 'The model %r has %d input(s) (%s). "pg" can only ' "execute models without inputs." % (
                filename,
                len(generator.input),
                inputs,
            )
            # XXX nothing given
            raise SemanticError(msg, generator.input[0])

        if debug:
            print(("Parsed model:\n\n%s" % generator.parsed_model))

        model = library.instance(filename, name="cmdline", config=config)

    else:
        # See if it exists
        if not os.path.exists(filename):
            # Maybe try with extension .pg
            filename_pg = "%s.pg" % filename
            if os.path.exists(filename_pg):
                filename = filename_pg
            else:
                # TODO: add where for command line
                msg = "Unknown model or file %r." % filename
                raise SemanticError(msg, None)  # XXX

        # Make sure we use absolute pathnames so that we know the exact
        # directory
        filename = os.path.realpath(filename)
        model_spec = open(filename).read()
        model = model_from_string(model_spec, config=config, filename=filename, library=library)

    if debug:
        model.summary()
        return

    # This is the main loop; unimpressive?

    count = 0
    try:
        model.init()

        while model.has_more():
            try:
                model.update()
            except BadMethodCall as e:
                e.blocks.insert(0, model)
                raise e

            if stats and stats_display_interval is not None:
                if count % stats_display_interval == 0:
                    s = model.stats_get_all()
                    y = yaml.dump(s)
                    with open("stats.yaml", "w") as f:
                        f.write(y)
                    model.print_stats()

                count += 1

        # info('Execution of %s finished quietly.' % model)
        model.finish()

    except KeyboardInterrupt:
        where = traceback.format_exc()
        error("Execution of %s interrupted by user at %s:" % (model, where))
        error("I will attempt clean-up.")
        raise
    except Exception as e:
        error("Execution of %s failed: %s" % (model, e))
        error("I will attempt clean-up.")
        raise
    finally:
        # info('Cleaning up.')
        model.cleanup()

    return model



import operator


__all__ = ["levenshtein", "levenshtein_best_match"]

# http://hetland.org/coding/python/levenshtein.py
def levenshtein(a, b):
    "Calculates the Levenshtein distance between a and b."
    n, m = len(a), len(b)
    if n > m:  # pragma: no cover
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n

    current = list(range(n + 1))
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)

    return current[n]


def levenshtein_best_match(s, options):
    matches = [(x, levenshtein(s[: len(x)], x) - len(x)) for x in options]
    best = min(matches, key=operator.itemgetter(1))
    return best



from . import np

from procgraph import simple_block
from procgraph.block_utils import assert_rgb_image


@simple_block
def posterize(rgb, levels=2):
    """ 
        Posterizes the given image with the specified number of levels.
        
        :param rgb: RGB image
        :type rgb: array(HxWx3,uint8),H>0,W>0
    
        :param levels: number of levels
        :type levels: int,>=2
        
        :return: A RGB image with the specified number of levels.
        :rtype: array(HxWx3,uint8)
    """
    assert_rgb_image(rgb, "input to rgb2grayscale")

    result = np.zeros(shape=rgb.shape, dtype="uint8")
    for i in range(3):
        channel = rgb[:, :, i].squeeze()
        result[:, :, i] = posterize_channel(channel, levels)

    return result


def posterize_channel(chan, levels):
    """ 
        Posterizes one channel of an image.
        
        :param rgb: image channel
        :type rgb: array(HxW,uint8),H>0,W>0
    
        :param levels: number of levels
        :type levels: int,>=2
        
        :return: Posterized channel
        :rtype: array(HxW,uint8)
    """

    assert levels > 1
    # Get the limits
    perc = np.linspace(0, 255.0, levels + 1)
    assert len(perc) == levels + 1

    result = np.zeros(shape=chan.shape, dtype="uint8")
    for level in range(levels):
        lower_bound = perc[level]
        upper_bound = perc[level + 1]
        assert lower_bound < upper_bound
        select = np.logical_and(chan >= lower_bound, chan <= upper_bound)
        target = (lower_bound + upper_bound) * 0.5
        result[select] = target
    return result



from procgraph import Block
from . import np  # @PydevCodeAnalysisIgnore


class Minimum(Block):
    """ Computes the minimum of a signal over time. """

    Block.alias("minimum_over_time")

    Block.input("x", "Any numpy array.")
    Block.output("min_x", "Minimum of input.")

    def init(self):
        self.state.min_x = None

    def update(self):
        # TODO: check shape did not change
        if self.state.min_x is None:
            self.state.min_x = self.input.x
        else:
            self.state.min_x = np.minimum(self.state.min_x, self.input.x)
        self.output.min_x = self.state.min_x


class Maximum(Block):
    """ Computes the minimum of a signal over time. """

    Block.alias("maximum_over_time")

    Block.input("x", "Any numpy array.")
    Block.output("max_x", "Maximum of input over time.")

    def init(self):
        self.state.max_x = None

    def update(self):
        # TODO: check shape did not change
        if self.state.max_x is None:
            self.state.max_x = self.input.x
        else:
            self.state.max_x = np.maximum(self.state.max_x, self.input.x)
        self.output.max_x = self.state.max_x



from itertools import tee

from geometry import SE2, SE3, se3, linear_angular_from_se2
import numpy as np
from procgraph import Block


class SE2_relative_pose(Block):
    """ Lets the first pose be the identity """

    Block.input("pose")
    Block.output("rel_pose")

    def init(self):
        self.state.pose0 = None

    def update(self):
        pose = self.input.pose
        if self.state.pose0 is None:
            self.state.pose0 = pose

        rel_pose = SE2.multiply(SE2.inverse(self.state.pose0), pose)

        self.output.rel_pose = rel_pose


class se2_from_SE2_seq(Block):
    """ Computes velocity in se2 given poses in SE2. """

    Block.alias("se2_from_SE2_seq")

    Block.input("pose", "Pose as an element of SE2", dtype=SE2)
    Block.output("velocity", "Velocity as an element of se(2).", dtype=np.dtype(("float", (3, 3))))

    def init(self):
        self.state.prev = None

    def update(self):
        q2 = self.get_input(0)
        t2 = self.get_input_timestamp(0)

        if self.state.prev is not None:
            t1, q1 = self.state.prev
            vel = velocity_from_poses(t1, q1, t2, q2)
            self.set_output(0, vel, timestamp=t2)

        self.state.prev = t2, q2


class vel_from_SE2_seq(Block):
    """ Computes velocity in se2 represented as vx,vy,omega
        from a sequence of poses in SE2. """

    Block.alias("vel_from_SE2_seq")

    Block.input("pose", "Pose as an element of SE2", dtype=SE2)
    Block.output("velocity", "Velocity as vx,vy,omega.", dtype=np.dtype(("float", 3)))

    def init(self):
        self.state.prev = None

    def update(self):
        q2 = self.get_input(0)
        t2 = self.get_input_timestamp(0)

        if self.state.prev is not None:
            t1, q1 = self.state.prev
            vel = velocity_from_poses(t1, q1, t2, q2, S=SE2)

            # Convertion from se2 to R3
            v, omega = linear_angular_from_se2(vel)
            out = np.array([v[0], v[1], omega])

            self.set_output(0, out, timestamp=t2)

        self.state.prev = t2, q2


class vel_from_SE3_seq(Block):
    """ Computes velocity in se3 represented as vx,vy,vz,w1,w2,w3
        from a sequence of poses in SE3. """

    Block.alias("vel_from_SE3_seq")

    Block.input("pose", "Pose as an element of SE3", dtype=SE3)
    Block.output("velocity", "Velocity as vx,vy,vz,w1,w2,w3", dtype=np.dtype(("float", 6)))

    def init(self):
        self.state.prev = None

    def update(self):
        q2 = self.get_input(0)
        t2 = self.get_input_timestamp(0)

        if self.state.prev is not None:
            t1, q1 = self.state.prev
            vel = velocity_from_poses(t1, q1, t2, q2, S=SE3)
            out = se3.vector_from_algebra(vel)
            self.set_output(0, out, timestamp=t2)

        self.state.prev = t2, q2


def pose_difference(poses, S=SE2):
    """ poses: sequence of (timestamp, pose) """

    for p1, p2 in pairwise(poses):
        t1, q1 = p1
        t2, q2 = p2
        v = velocity_from_poses(t1, q1, t2, q2, S=S)
        yield v


def velocity_from_poses(t1, q1, t2, q2, S=SE2):
    delta = t2 - t1
    if not delta > 0:
        raise ValueError("invalid sequence")

    x = S.multiply(S.inverse(q1), q2)
    xt = S.algebra_from_group(x)
    v = xt / delta
    return v


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return list(zip(a, b))


#
# class MyInput(object):
#     def __init__(self, block):
#         self.block = block
#
#
# class SimpleSequenceBlock(IteratorGenerator):
#
#     def init_iterator(self):
#         my_input = MyInput(self)
#         myit = self.get_iterator(my_input)
#
#         def add_signal_name(it):
#             for timestamp, value in it:
#                 yield timestamp, 0, value
#
#         return add_signal_name(myit)
#
#     @abstractmethod
#     @contract(returns=GeneratorType)
#     def get_iterator(self, input_sequence):
#         pass
#
#     def init(self):
#         self.iterator = self.init_iterator()
#         if self.iterator is None:
#             msg = 'must return an iterator, got %s' % describe_value(self.iterator)
#             raise ValueError(msg)
#         self._load_next()
#
#     def _load_next(self):
#         try:
#             signal, timestamp, value = self.iterator.next()
#             if not isinstance(signal, (str, int)):
#                 msg = ('Expected a string or number for the signal, got %s' %
#                        describe_value(signal))
#                 raise ValueError(msg)
#             if not isinstance(timestamp, float):
#                 msg = ('Expected a number for the timestamp, got %s' %
#                        describe_value(timestamp))
#                 raise ValueError(msg)
#
#             self.next_signal = signal
#             self.next_timestamp = timestamp
#             self.next_value = value
#             self.has_next = True
#         except StopIteration:
#             self.has_next = False
#
#     def next_data_status(self):
#         if self.has_next:
#             return (True, self.next_timestamp)
#         else:
#             return (False, None)
#
#     def update(self):
#         if not self.has_next:
#             return  # XXX: error here?
#
#         self.set_output(self.next_signal,
#                         value=self.next_value, timestamp=self.next_timestamp)
#
#         self._load_next()
#
#
# class VelocityFromPoses(SimpleSequenceBlock):
#
#     Block.input('pose', 'Pose as an element of SE2')
#
#     def get_iterator(self, input_sequence):
#         for t, x in pose_difference(input_sequence):
#             yield t, x
#
#



""" 
    Blocks for basic operations on images.

    This package contains blocks that perform basic operations
    on images. The library has no software dependency. 
    
    For more complex operations see also :ref:`module:procgraph_cv` and 
    :ref:`module:procgraph_pil`
    
    **Example**
    
    Convert a RGB image to grayscale, and back to a RGB image:::
    
    
        |input| -> |rgb2gray| -> |gray2rgb| -> |output| 

"""
from procgraph import import_magic, import_successful


procgraph_info = {
    "requires": ["numpy"],
}


np = import_magic(__name__, "numpy")


from .filters import *
from .compose import *
from .imggrid import *
from .images_from_dir import *
from .posterize import *

from .border import *
from .pad import *
from .blend import *
from .copied_from_reprep import *

from .reshape_smart import *
from .solid_imp import *
from .crop import *

from .alpha import *

from .reshape_rect import *



from procgraph import Block, COMPULSORY, simple_block
import numpy as np


class Extract(Block):
    """ 
    This block extracts some of the components of a vector.
    
    Example: Extracts the first and third component of x. ::
    
        x -> |extract index=[0,2]| -> x_part
        
        
    """

    Block.alias("extract")
    Block.input("vector", "Any numpy array")
    Block.output("part", "The part extracted")
    Block.config("index", "Index (or indices) to extract.")

    def update(self):
        index = self.config.index
        vector = np.array(self.input.vector)

        part = vector[index]

        self.output.part = part


@simple_block
def slice(signal, start=COMPULSORY, end=COMPULSORY):  # @ReservedAssignment
    """ Slices a signal by extracting from index ``start`` to index ``end``
        (INCLUSIVE).
        
        :param signal: Any 1d numpy array
        :param start:  Slice start.
        :param end:    Slice end (inclusive).
        
        :return: sliced: The sliced signal.
    """
    assert start != COMPULSORY
    assert end != COMPULSORY
    return signal[start : (end + 1)]



from contracts import contract, describe_value
from procgraph import BadInput, simple_block
from procgraph.block_utils import assert_gray_image, assert_rgb_image
from . import np


@simple_block
def as_uint8(rgb):
    res = rgb.astype("uint8")
    return res


@simple_block
def as_float32(rgb):
    res = rgb.astype("float32")
    return res


@simple_block
@contract(x="array[HxW](>=0, <=255)", returns="array[HxW](>=0, <=1)")
def range_01_from_255(x):
    """ Divides by 255. """
    return x / 255.0


@simple_block
@contract(returns="array[HxWx3](uint8)")
def torgb(rgb):
    """ Converts all image formats to RGB uint8. """

    try:
        # check_convertible_to_rgb(rgb)
        pass
    except ValueError as e:
        raise BadInput(str(e), None, 0)

    if rgb.ndim == 2:
        if rgb.dtype == "float32":
            m = np.amin(rgb)
            M = np.amax(rgb)
            if m < 0 or M > 1:
                msg = "If float, I assume range [0,1]; found [%s, %s]." % (m, M)
                raise ValueError(msg)
            rgb = (rgb * 255).astype("uint8")
        return gray2rgb(rgb)

    nc = rgb.shape[2]
    if nc == 4:
        return rgb[:, :, :3]

    if nc == 3:
        if rgb.dtype == "float32":
            return (rgb * 255).astype("uint8")
        elif rgb.dtype == "uint8":
            return rgb
        else:
            msg = "Expected image format: %s" % describe_value(rgb)
            raise ValueError(msg)


@simple_block
@contract
def rgb2gray(rgb):
    """ Converts a HxWx3 RGB image into a HxW grayscale image 
        by computing the luminance. 
        
        :param rgb: RGB image
        :type rgb: array[HxWx3](uint8),H>0,W>0
        
        :return: A RGB image in shades of gray.
        :rtype: array[HxW](uint8)
    """
    assert_rgb_image(rgb, "input to rgb2grayscale")
    r = rgb[:, :, 0].squeeze()
    g = rgb[:, :, 1].squeeze()
    b = rgb[:, :, 2].squeeze()
    # note we keep a uint8
    gray = r * 299.0 / 1000 + g * 587.0 / 1000 + b * 114.0 / 1000
    gray = gray.astype("uint8")

    return gray


@simple_block
@contract
def gray2rgb(gray):
    """ 
        Converts a H x W grayscale into a H x W x 3 RGB image 
        by replicating the gray channel over R,G,B. 
        
        :param gray: grayscale
        :type  gray: array[HxW](uint8),H>0,W>0
        
        :return: A RGB image in shades of gray.
        :rtype: array[HxWx3](uint8)
    """
    assert_gray_image(gray, "input to gray2rgb")

    rgb = np.zeros((gray.shape[0], gray.shape[1], 3), dtype="uint8")
    for i in range(3):
        rgb[:, :, i] = gray
    return rgb



from .containers import CONTAINERS, CONTAINER_MP4, do_quickstart, guess_container, supports_full_metadata
from .metadata import get_ffmpeg_metadata_args, write_extra_metadata_for
from .vcodecs import VCODECS, guess_vcodec
from .video_info import pg_video_info
from contracts import contract
from procgraph import logger
from procgraph.utils import CmdException, system_cmd_result
import os


__all__ = ["pg_video_convert"]


@contract(timestamp="None|float|int", metadata="dict", vcodec_params="dict")
def pg_video_convert(
    filename, out, quiet=True, container=None, vcodec=None, vcodec_params={}, timestamp=None, metadata={}
):
    """
        Converts a video file (e.g. an AVI) to another format.
        
        
        container: one of "mkv", "mp4", "mov"
        
        vcodec : vcodec params with defaults
        'prores': profile=3, qv=None
        'x264': crf=18, preset='medium'
        
        It makes sure to write information to preserve timestamp 
        and the given metadata.
        
        One can then be guaranteed to access this data using 
        the pg_info_video() function. 
    """
    logger.info("pg_video_convert:\n<- %s\n-> %s" % (filename, out))

    if container is None:
        container = guess_container(out)

    assert container in CONTAINERS

    if vcodec is None:
        vcodec, vcodec_params = guess_vcodec(container)

    logger.info("container: %s" % container)
    logger.info("vcodec: %s" % vcodec)
    logger.info("vcodec_params: %s" % vcodec_params)

    no_audio = True

    cmds = ["ffmpeg"]

    # For using aac out
    # cmds += ['-strict', '-2']

    cmds += ["-y"]
    cmds += ["-i", filename]

    if no_audio:
        cmds += ["-an"]

    cmds += VCODECS[vcodec](**vcodec_params)

    info = pg_video_info(filename, intolerant=False)
    info["metadata"].update(metadata)

    if timestamp is None:
        timestamp = info["timestamp"]
    else:
        timestamp = float(timestamp)

    cmds += get_ffmpeg_metadata_args(metadata, timestamp)
    # cmds += ['-f', container]

    if container == CONTAINER_MP4:
        out1 = out + ".firstpass.mp4"
        cmds += [out1]
    else:
        out1 = out
        cmds += [out1]

    try:
        system_cmd_result(
            ".",
            cmds,
            display_stdout=not quiet,
            display_stderr=not quiet,
            raise_on_error=True,
            capture_keyboard_interrupt=False,
        )
    except CmdException:
        if os.path.exists(out1):
            os.unlink(out1)
        raise

    assert os.path.exists(out1)

    if container == CONTAINER_MP4:
        do_quickstart(out1, out)
        os.remove(out1)
        # warnings.warn("Not sure why quickstart does not work.")
        # os.rename(out1, out)
    else:
        assert out1 == out

    if not os.path.exists(out):
        logger.error("Something is wrong, path does not exist.\nPath: %s" % out)

    if not supports_full_metadata(container):
        write_extra_metadata_for(out, metadata)



#!/usr/bin/env python
from optparse import OptionParser
from procgraph import pg, register_model_spec
import os

usage = """
    %cmd  -o <output> -p <pattern> -d <dirname>
"""


def main():
    parser = OptionParser(usage=usage)

    parser.add_option("-o", dest="output", type="string", help="Output video.")

    parser.add_option("-d", dest="dirname", type="string", default=".", help="Directory.")

    parser.add_option("--fps", type="float", default=10.0, help="Frames per second")

    parser.add_option("-p", dest="pattern", type="string", default="([\w-]+)(\d+)\.(\w+)", help="Pattern")

    (options, args) = parser.parse_args()
    if args:
        raise Exception("Spurious arg: %s" % args)

    if options.output is None:
        options.output = os.path.join(options.dirname, "output.mp4")

    return join_video(options.output, options.dirname, options.pattern, options.fps)


def join_video(output, dirname, pattern, fps):
    """ 
        Note that the pattern is a Python regex:
        
        pg-video-join -d anim-simple/ -p '*.png' -o anim-simple-collate.mp4 --fps 1 
        
    """
    register_model_spec(
        """
--- model join_video_helper
config output
config dirname
config pattern
config fps

|files_from_dir dir=$dirname regexp=$pattern| -->file
#file --> |print|
file --> |imread_rgb| --> rgb
#file --> |imread| --> |torgb| --> rgb


rgb --> |mencoder quiet=1 file=$output timestamps=0 fps=$fps|
    
    """
    )

    pg("join_video_helper", dict(dirname=dirname, pattern=pattern, output=output, fps=fps))


def join_video_29(output, dirname, pattern, fps):
    """ 
        Note that the pattern is a Python regex:
        
        pg-video-join -d anim-simple/ -p '*.png' -o anim-simple-collate.mp4 --fps 1 
        
    """
    register_model_spec(
        """
--- model join_video_helper_29
config output
config dirname
config pattern
config factor

|files_from_dir dir=$dirname regexp=$pattern| -->file
file --> |imread_rgb| --> rgb
rgb --> |rewrite_timestamps interval=$factor| -> retimed
retimed --> |info|
# XXX this doesn't work
#retimed -> |fix_frame_rate fps=29.97| -->  fill
#fill --> |info|
retimed -> |mencoder quiet=1 file=$output timestamps=0|
    
    """
    )
    factor = 1.0 / fps
    pg("join_video_helper_29", dict(dirname=dirname, pattern=pattern, output=output, factor=factor))


if __name__ == "__main__":
    main()



from .border import image_border


__all__ = ["image_pad"]


def image_pad(pixel_data, expected_shape, bgcolor=[1, 1, 1]):
    """ Pads horizontally and vertically an image so that it is of the 
        given shape. """
    shape = pixel_data.shape[0:2]
    if (shape[0] > expected_shape[0]) or (shape[1] > expected_shape[1]):
        msg = "Image is bigger (%s) than the expected shape (%s) " % (shape, expected_shape)
        raise ValueError(msg)

    extra_hor = expected_shape[1] - shape[1]
    left = int(extra_hor / 2)
    right = extra_hor - left

    extra_ver = expected_shape[0] - shape[0]
    top = int(extra_ver / 2)
    bottom = extra_ver - top

    new_image = image_border(rgb=pixel_data, color=bgcolor, left=left, right=right, top=top, bottom=bottom)

    # assert_equal(new_image.shape[0:2], expected_shape)

    return new_image



from .utils import PGTestCase


class TestRenaming(PGTestCase):
    def test_renaming(self):
        """ Sometimes instancing a block twice will give error because
            we overwrote the parsed spec. """
        model_spec = """
--- model loaders

|input name=y|  --> |identity| --> y_dot
y_dot  --> |identity| --> y_ddot
 
         """
        self.check_semantic_ok(model_spec)



from procgraph import Block


class Any(Block):
    """ 
        Joins the stream of multiple signals onto one output signal.
    """

    Block.alias("any")

    Block.input_is_variable("Signals to be put on the same stream.", min=1)
    Block.output("stream", "Unified stream.")

    def init(self):
        self.last_ts = {}
        self.buffer = []

    def update(self):
        # DO NOT USE NAMES
        # TODO: check other blocks for this bug
        nsignals = self.num_input_signals()
        for i in range(nsignals):
            value = self.get_input(i)
            ts = self.get_input_timestamp(i)
            if value is None:
                continue
            if not i in self.last_ts or ts != self.last_ts[i]:
                self.buffer.append((ts, value))
            self.last_ts[i] = ts

        #        t = [x[0] for x in self.buffer]
        #        self.debug(' after1: %s' % t)
        #
        self.buffer = sorted(self.buffer, key=lambda x: x[0])

        #        t = [x[0] for x in self.buffer]
        #        self.debug(' sort: %s' % t)

        # Make sure we saw every signal before outputing one
        #        if len(self.last_ts) == nsignals:
        # FIXME: bug we will send one sample of the last stream

        if self.buffer:
            ts, value = self.buffer.pop(0)
            self.set_output(0, value, ts)


#        t = [x[0] for x in self.buffer]
#        self.debug(' after2: %s' % t)

#        self.debug('Got %s, chose %s' % (),
#                                                timestamp))
#



import numpy

from procgraph import simple_block

# Taken from the numpy cookbook


@simple_block
def smooth1d(x, window_len=11, window="hanning"):
    """
    Smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    ``window`` must be one of  'flat', 'hanning', 'hamming', 'bartlett', 
    'blackman'.
    A flat window will produce a moving average smoothing.

    
    :param x: the input signal 
    :param window_len: the dimension of the smoothing window;  an odd integer
    :param window: the type of window from 
    :return: smoothed: the smoothed signal
        
    example: ::

        t=linspace(-2,2,0.1)
        x=sin(t)+randn(len(t))*0.1
        y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, 
    numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an 
          array instead of a string   
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            "Window %r is not one of 'flat', 'hanning', 'hamming'," "'bartlett', 'blackman'." % window
        )

    s = numpy.r_[2 * x[0] - x[window_len:1:-1], x, 2 * x[-1] - x[-1:-window_len:-1]]
    # print(len(s))
    if window == "flat":  # moving average
        w = numpy.ones(window_len, "d")
    else:
        w = eval("numpy." + window + "(window_len)")

    y = numpy.convolve(w / w.sum(), s, mode="same")
    return y[window_len - 1 : -window_len + 1]



from procgraph.core.registrar_other import simple_block
from contracts import contract
import numpy as np


@simple_block
@contract(bg="array[HxWx(3|4)]", moving="array[HxWx4]")
def alpha_add(moving, bg):
    H, W = bg.shape[0:2]
    res = np.zeros((H, W, 3), "uint8")

    im1 = bg
    im2 = moving
    alpha2 = moving[:, :, 3]
    #     alpha2 = alpha2 / 255.0 #np.max(alpha2)
    alpha2 = alpha2 > 0
    alpha1 = 1.0 - alpha2

    for i in range(3):
        res[:, :, i] = im1[:, :, i] * alpha1 + im2[:, :, i] * alpha2

    return res



from procgraph import Block, BadInput, register_model_spec

from . import geometry

g = geometry


class Pose2velocity(Block):
    """ Block used by :ref:`block:pose2commands`. """

    Block.alias("pose2vel_")

    Block.input("q12", "Last two poses.")
    Block.input("t12", "Last two timestamps.")

    Block.output("commands", "Estimated commands ``[vx,vy,omega]``.")

    def update(self):
        q = self.get_input("q12")
        t = self.get_input("t12")
        if not (len(q) == 2 and len(t) == 2):
            raise BadInput("Bad input received.", self, q)

        pose1 = g.SE2_from_xytheta(q[0])
        pose2 = g.SE2_from_xytheta(q[1])
        delta = t[1] - t[0]

        if not delta > 0:
            raise BadInput("Bad timestamp sequence %s" % t, self, "t")

        _, vel = g.SE2.velocity_from_points(pose1, pose2, delta)

        linear, angular = g.linear_angular_from_se2(vel)

        commands = [linear[0], linear[1], angular]
        self.set_output("commands", commands, timestamp=t[0])


# TODO: move this to models/
register_model_spec(
    """
--- model pose2commands
''' Computes the velocity commands from the odometry data. '''

input pose "Odometry as an array ``[x,y,theta]``."

output commands "Estimated commands as an array ``[vx,vy,omega]``."
output vx       "Linear velocity, forward (m/s)"
output vy       "Linear velocity, side (m/s)"
output omega    "Angular velocity (rad/s)"
 
|input name=pose| --> |last_n_samples n=2| --> |pose2vel_| --> commands 

    commands          -->          |output name=commands|
    commands --> |extract index=0| --> |output name=vx|
    commands --> |extract index=1| --> |output name=vy|
    commands --> |extract index=2| --> |output name=omega|
    
"""
)


register_model_spec(
    """
--- model commands_from_SE2
''' Computes the velocity commands from the odometry data. '''

input pose "Odometry as an element of SE2."

output commands "Estimated commands as an array ``[vx,vy,omega]``."
output vx       "Linear velocity, forward (m/s)"
output vy       "Linear velocity, side (m/s)"
output omega    "Angular velocity (rad/s)"
 
|input name=pose| --> |last_n_samples n=2| --> |pose2vel_| --> commands 

    commands          -->          |output name=commands|
    commands --> |extract index=0| --> |output name=vx|
    commands --> |extract index=1| --> |output name=vy|
    commands --> |extract index=2| --> |output name=omega|
    
"""
)



from procgraph import Block

__all__ = ["Wait"]


class Wait(Block):
    """ 
        This block waits a given number of updates before transmitting the 
        output signal.
    """

    Block.alias("wait")

    Block.config("n", "Number of updates to wait at the beginning.")

    Block.input_is_variable("Arbitrary signals.")
    Block.output_is_variable("Copy of the signals, minus the first " "*n* updates.")

    def init(self):
        self.state.count = 0

    def update(self):
        count = self.state.count

        # make something happen after we have waited enough
        if count >= self.config.n:
            # Just copy the input to the output
            for i in range(self.num_input_signals()):
                self.set_output(i, self.get_input(i), self.get_input_timestamp(i))

        self.state.count = count + 1



""" Components used for debugging and unit tests. """

from . import constant
from . import identity
from . import info
from . import printc
from . import gain
from . import operations
from . import clock



# from .utils import PGTestCase


# class SemanticsTest(PGTestCase):
#
#     def testBadExamples(self):
#         for example in bad_examples2:
#             self.check_semantic_error(example)
#
#     def testExamples(self):
#         for example in good_examples2:
#             self.check_semantic_ok(example)
#
#
# class ParamsTest(PGTestCase):
#
#     def test1(self):
#         ''' Checking that we signal unused parameters. '''
#
#         model_desc = """
#            |input name=x| -> |g1:gain| -> |output name=y|
#         """
#         # g2.gain is unused
#         self.assertRaises(SemanticError, model_from_string,
#                           model_desc, config={'g2.k': 2})
#



import inspect
import types

from .block import Block
from .constants import COMPULSORY, TIMESTAMP
from .constants import NO_OUTPUT
from .docstring_parsing import parse_docstring_annotations, DocStringInfo
from .exceptions import BadConfig, BadInput
from .model_loader import add_models_to_library
from .registrar import default_library


def make_generic(name, inputs, num_outputs, operation, params={}, docs=None):
    # XXX: This is not pickable
    # make a copy
    parameters = dict(params)

    if docs is None:
        docstring = operation.__doc__
    else:
        docstring = docs

    try:
        if docstring is not None:
            annotations = parse_docstring_annotations(docstring)
            docstring = annotations.docstring
        else:
            annotations = DocStringInfo(docstring)
    except Exception as e:
        # print('Malformed annotation for %r: %s' % (operation, e))
        # print docstring
        raise

    def get_param_annotation(key):
        if key in annotations.params:
            arg = annotations.params[key]
            description = arg.desc
            if description is None:
                description = ""
            if arg.type is not None:
                description += " (%s)" % arg.type

        else:
            description = None
        return description

    class GenericOperation(Block):
        Block.alias(name)

        # filled out later
        defined_in = None

        # XXX: does it work this way?
        __doc__ = docstring  # @ReservedAssignment
        my_operation = operation

        for key, value in list(parameters.items()):
            description = get_param_annotation(key)

            if not value in [TIMESTAMP]:
                if value == COMPULSORY:
                    Block.config(key, description=description)
                else:
                    Block.config(key, description=description, default=value)

        for input_signal in inputs:
            description = get_param_annotation(input_signal)
            Block.input(input_signal, description=description)

        for i in range(num_outputs):
            output_name = str(i)
            if i < len(annotations.returns):
                arg = annotations.returns[i]
                description = arg.desc
                if description is None:
                    description = ""
                if arg.type is not None:
                    description += " (%s)" % arg.type
                tokens = description.split(":")
                if len(tokens) == 2:
                    output_name = tokens[0]
                    description = tokens[1]
                    # TODO: check good name
            else:
                description = None

            Block.output(output_name, description=description)

        def update(self):
            args = []
            for input_signal in inputs:
                args.append(self.get_input(input_signal))

            params = {}
            for key in parameters:
                if parameters[key] == TIMESTAMP:
                    params[key] = max(self.get_input_signals_timestamps())
                else:
                    params[key] = self.get_config(key)

            try:
                result = operation(*args, **params)
            # Functions can throw BadInput and BadConfig, but we have to fill
            # in the block reference for them.
            except (BadInput, BadConfig) as e:
                e.block = self
                raise e

            no_out = isinstance(result, str) and str == NO_OUTPUT

            if not no_out:
                if num_outputs == 1:
                    self.set_output(0, result)
                else:
                    for i in range(num_outputs):
                        self.set_output(i, result[i])

    return GenericOperation


# TODO: add num_inputs, so that we can get rid of COMPULSORY.
def simple_block(alias=None, num_outputs=1):
    """ Decorator for turning functions into simple blocks. """

    # OK, this is black magic. You are not expected to understand this.
    if type(alias) is types.FunctionType:
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        defined_in = mod.__name__

        function = alias
        alias = None
        num_outputs = 1
        # print "Registering %s: %s (no params) " % (defined_in, function)
        register_simple_block(function, name=alias, num_outputs=num_outputs, defined_in=defined_in)
        return function
    else:

        def wrap(function):
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            defined_in = mod.__name__

            # print("Registering %s: %s (with params %s %s)" %
            #      (defined_in, function, alias, num_outputs))
            register_simple_block(function, name=alias, num_outputs=num_outputs, defined_in=defined_in)
            return function

        return wrap


def register_simple_block(
    function, name=None, num_inputs=1, num_outputs=1, params={}, doc=None, defined_in=None
):
    # Get a module to which we can associate this block
    if defined_in is None:
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        defined_in = mod.__name__

    try:
        args, _, _, defaults = inspect.getargspec(function)
        # TODO: use varwk for variable signals
        num_defaults = len(defaults) if defaults else 0
        num_no_argument = len(args) - num_defaults
        args_no_argument = args[:num_no_argument]
        args_with_default = args[num_no_argument:]
        config = dict([(args_with_default[i], defaults[i]) for i in range(num_defaults)])
        inputs = args_no_argument
    except Exception as e:  # @UnusedVariable
        # TODO: add switch to show this
        #         print("Does not work with %s: %s " % (function, e))
        config = params
        inputs = [str(i) for i in range(num_inputs)]

    assert name is None or isinstance(name, str)
    if name is None:
        name = function.__name__

    block = make_generic(name, inputs, num_outputs, function, params=config, docs=doc)

    block.defined_in = defined_in

    assert isinstance(block.defined_in, str)


def register_block(block_class, name=None):
    assert name is None or isinstance(name, str)
    if name is None:
        name = block_class.__name__
    default_library.register(name, block_class)


already = set()


def register_model_spec(model_spec, defined_in=None):
    if model_spec in already:
        return
    already.add(model_spec)
    if defined_in is None:
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        if mod is None:
            msg = "Could not find module name.\nfrm[0]=%r" % frm[0]
            raise ValueError(msg)
        defined_in = mod.__name__
    add_models_to_library(default_library, model_spec, defined_in=defined_in)



""" ProcGraph: what you would get if Simulink was written in Python
    and was actually useful for dealing with log data. """

__version__ = "6.1.1"

from zuper_commons.logs import ZLogger

logger = ZLogger(__name__)
import os

path = os.path.dirname(os.path.dirname(__file__))
logger.info(f"procgraph version {__version__} path {path}")

# If true, does not allow .pgc caches
deny_pgc_cache = False

from .core.exceptions import *

from .core.block import Block, Generator
from .core.model import Model

from .core.model_loader import pg_add_this_package_models

from .core.registrar_other import register_model_spec, register_simple_block, simple_block

from .core.registrar import Library, default_library

from .scripts import pg

from .core.import_magic import import_magic, import_succesful, import_successful

from .core.constants import *

from .block_utils import IteratorGenerator

from .components import *



from ..core.model_loader import model_from_string

from .utils import PGTestCase

examples = [
    """ # direct
   |input name=x| -> |g1:gain| -> |output name=y|
""",
    """ # named
   |input name=x| -> v -> |g1:gain| -> |output name=y|
""",
    """ # named
   |input name=x| -> v -> |g1:gain| -> y -> |output name=y|
""",
    """ # only  identity 2
   |input name=x| -> |identity| -> |g1:gain| -> |identity| -> |output name=y|
""",
    """ # named + identity
   |input name=x| -> v -> |identity| -> |g1:gain| -> y -> |output name=y|
""",
    """ # named + identity
   |input name=x| -> |identity| -> v -> |g1:gain| -> y -> |output name=y|
""",
    """ # only  identity
   |input name=x| -> |identity| -> |g1:gain| -> y -> |output name=y|
""",
    """
   (x) -> |g1:gain| -> (y)
""",
]


class PipelineTest(PGTestCase):
    def test_pipeline(self):
        """ All graphs equivalent (multiply by a gain), but interconnections
            change. """
        for example in examples:
            self.try_one(example)

    def try_one(self, model_desc):
        print(("Trying with '''%s'''" % model_desc))
        gain = 3
        model = model_from_string(model_desc, config={"g1.k": gain})
        model.init()

        for i in range(5):
            value = i
            timestamp = 4 + i * 0.5
            model.from_outside_set_input("x", value, timestamp=timestamp)

            self.assertTrue(model.has_more())
            while model.has_more():
                print(("update %s  y = %s" % (i, model.get_output(0))))
                model.update()

            self.assertEqual(model.get_output(0), gain * value)
            self.assertEqual(model.get_output_timestamp(0), timestamp)



from procgraph.core.registrar import default_library, Library
from procgraph.core.model_loader import model_from_string

from .utils import PGTestCase

# All of these models, when interpreted,
# have an output that equals 42

examples_42 = [
    (
        """
--- model master
# first
output meaning

|slave my_param=42| ---> |output name=meaning|

--- model slave
config my_param
output c

|c1:constant| --> |output name=c|

c1.value = $my_param

""",
        {},
    ),
    (
        """
--- model master
output meaning

|slave my_param=$my_param2| ---> |output name=meaning|

my_param2 = 42

--- model slave
config my_param
output c

|c1:constant| --> |output name=c|

c1.value = $my_param

""",
        {},
    ),
    (
        """
--- model master
output meaning

|s1:slave| ---> |output name=meaning|

s1.my_param = 42

--- model slave
config my_param
output c

|c1:constant| --> |output name=c|

c1.value = $my_param

""",
        {},
    ),
    (
        """
--- model master
my_param2 = 42
output meaning

|s1:slave| ---> |output name=meaning|

s1.my_param = $my_param2

--- model slave
config my_param
output c

|c1:constant| --> |output name=c|

c1.value = $my_param

""",
        {},
    ),
    (
        """
--- model master
config my_param2
output meaning

|s1:slave| ---> |output name=meaning|

s1.my_param = $my_param2

--- model slave
config my_param
output c
|c1:constant| --> |output name=c|

c1.value = $my_param

""",
        {"my_param2": 42},
    ),
    (
        """ # Default naming
--- model master
output meaning

|slave| ---> |output name=meaning|

--- model slave
config my_param
output c

|constant| --> |output name=c|

constant.value = $my_param

""",
        {"slave.my_param": 42},
    ),
    (
        """ # Default naming -- recursive.
--- model master
output meaning

|slave| ---> |output name=meaning|

--- model slave
output c

|constant| --> |output name=c|

""",
        {"slave.constant.value": 42},
    ),
]


class ParamsTest(PGTestCase):
    def test1(self):
        for model_spec, config in examples_42:
            # don't pollute the main library with the block definitions
            library = Library(default_library)
            model = model_from_string(model_spec, config=config, library=library)

            print(("Trying with '''%s'''" % model_spec))
            model.init()
            print("before")
            # at least the constant must be propagated
            assert model.has_more()

            while model.has_more():
                print(("before iteration", model.get_output(0)))
                model.update()
            model.finish()
            print(("end", model.get_output(0)))

            self.assertEqual(model.get_output(0), 42)



import sys
from types import ModuleType

from .visualization import debug
from .constants import PROCGRAPH_INFO_VARIABLE, REQUIRES, REQUIRES_PARSED


def import_magic(module_name, required, member=None):
    """ Equivalent to "from required import member" or "import required".
        Check that it was succesfull with import_succesful().
    """
    info_structure = get_module_info(module_name)
    # Check that it was mentioned in the structure
    required_base = required.split(".")[0]
    if not required_base in info_structure[REQUIRES_PARSED]:
        raise Exception(
            "Please specify that you need %r as a dependency "
            "in the field %r of the %r structure in module %r."
            % (required_base, REQUIRES, PROCGRAPH_INFO_VARIABLE, module_name)
        )

    # FIXME: there's a bug in here, should find which base was selected
    if required == required_base:
        options = info_structure[REQUIRES_PARSED][required]
    else:
        options = [required]

    errors = ""
    for o in options:
        try:
            mod = __import__(o, fromlist=["dummy"])
            if member is not None:
                if not member in mod.__dict__:
                    raise Exception("No member %r in %r" % (member, o))
                return mod.__dict__[member]
            return mod
        except Exception as e:  # @UnusedVariable
            # TODO: show error
            # debug(e)
            errors += "\n" + str(e)
            pass

    # We could not load anything.
    warn = (
        "Could not load dependency %r for %r.\n I will try to continue,"
        " but an error might be thrown when %s actually tries to use %r."
        % (required, module_name, module_name, required)
    )

    if False:  # TODO: think of configuration switch
        debug(warn)

    msg = "I tried to let you continue, but it seems that module %r really" " needs %r to work. Sorry! " % (
        module_name,
        required,
    )

    msg += "\nThe error message was:\n%s" % errors

    class warn_and_throw:
        def __getattr__(self, method_name):  # @UnusedVariable
            raise Exception(msg)

    return warn_and_throw()


def get_module_info(module_name):
    # TODO: better Exception?
    if not module_name in sys.modules:
        raise Exception("Please pass  __package__ as argument (got: %r)." % module_name)

    module = sys.modules[module_name]

    if not PROCGRAPH_INFO_VARIABLE in module.__dict__:
        raise Exception(
            "Please define the structure %r for module %r. " % (PROCGRAPH_INFO_VARIABLE, module_name)
        )

    info = module.__dict__[PROCGRAPH_INFO_VARIABLE]

    """ Returns dict   name -> list of possible modules """
    parsed = {}
    if not REQUIRES in info:
        # raise Exception('Please define a field %r in dict %s.%s.' %
        #                (REQUIRES, module_name, PROCGRAPH_INFO_VARIABLE))
        pass
    else:
        requires = info[REQUIRES]
        for r in requires:
            if isinstance(r, str):
                # normal
                parsed[r] = [r]
            else:
                # TODO: check iterable
                name = r[0]
                options = list(r[1])
                # TODO: check options > 0
                parsed[name] = options

    info[REQUIRES_PARSED] = parsed

    return info


def import_successful(m):
    return isinstance(m, ModuleType)


import_succesful = import_successful  # oops, typo



from procgraph import Generator, Block, ModelExecutionError, register_model_spec, logger
from procgraph.core.registrar import default_library
from procgraph.testing.utils import PGTestCase


examples = [
    (
        "If coincident, output that",
        """
1 2 3 4 5 6 7 8 9
1 2 3 4 5 6 7 8 9
1 2 3 4 5 6 7 8 9
""",
    ),
    (
        "Skip additional",
        """
1 2 3 4 5 6 7 8 9
1*2*3*4*5*6*7*8*9
1 2 3 4 5 6 7 8 9
""",
    ),
    (
        "Wait for the next.",
        """
1 2 3 4 5 6 7 8 9
 1 2 3 4 5 6 7 8 9
 1 2 3 4 5 6 7 8 9
""",
    ),
    (
        "Skip first one if necessary",
        """
 1 2 3 4 5 6 7 8 9
* 1 2 3 4 5 6 7 8 9
* 1 2 3 4 5 6 7 8 9
""",
    ),
    (
        "Wait for the next  (multiple delay).",
        """
1   2   3   4   5   6   7   8   9
 1   2   3   4   5   6   7   8   9
  1   2   3   4   5   6   7   8   9
""",
    ),
    (
        "Skip master if signals don't arrive .",
        """
1  *2  *3  **4***5**6   7   8   9
 1   2   3   4   5   6   7   8   9
 1   2   3   4   5   6   7   8   9
""",
    ),
    (
        "Skip slave if other slave doesn't arrive.",
        """
1   2   3   4   5   6   7   8   9
 1*  2   3   4   5   6   7   8   9
  1   2   3   4   5   6   7   8   9
""",
    ),
    (
        "Skip slave if other slave doesn't arrive.",
        """
* 1   2   3   4   9
 1   *   2   3   4   9
   1   2   3   4   9
""",
    ),
]

register_model_spec(
    """
--- model sync_test

|log_sim line=$master| -> master
|log_sim line=$slave1| -> slave1
|log_sim line=$slave2| -> slave2

master, slave1, slave2 -> |sync2| -> |check_sync| --> |output name=value|

"""
)


class LogSim(Generator):
    Block.alias("log_sim")

    Block.output("stream")

    def init(self):
        line = self.config.line

        self.queue = []
        for i, char in enumerate(line):
            t = i + 1
            if char == " ":
                continue
            elif char == "*":
                self.queue.insert(0, (t, -1))
            elif char.isdigit():
                v = int(char)
                self.queue.insert(0, (t, v))
            else:
                raise Exception('Bad input "%s" at char %d of "%s".' % (char, t, line))

    def next_data_status(self):
        if self.queue:
            return (True, self.queue[-1][0])
        else:
            return (False, None)

    def update(self):
        assert self.queue
        timestamp, value = self.queue.pop()
        self.set_output(0, value, timestamp)


class CheckSync(Block):
    Block.alias("check_sync")
    Block.output("value")

    def update(self):

        values = self.get_input_signals_values()
        timestamps = self.get_input_signals_timestamps()

        ok = True
        for v in values:
            if v is None or v != values[0]:
                ok = False
        for t in timestamps:
            if t is None or t != timestamps[0]:
                ok = False

        if not ok:
            msg = "Received bad inputs. Timestamps: %s Values: %s " % (timestamps, values)

            raise ModelExecutionError(msg, self)

        self.output.value = values[0]


class TestSync(PGTestCase):
    def not_working(self):

        for description, example in examples:
            logger.info("Trying with: %s" % description)
            # get non-empty lines
            lines = list([x for x in example.split("\n") if x.strip()])
            logger.info(lines=lines)
            assert len(lines) == 3
            config = {"master": lines[0], "slave1": lines[1], "slave2": lines[2]}

            model = default_library.instance("sync_test", name=None, config=config)

            model.init()
            while model.has_more():
                model.update()
            model.finish()

            final = model.get_output(0)
            if final != 9:
                raise Exception("expected final output to be 9, not %s" % final)



from procgraph import Block
import numpy as np


class TimeSlice(Block):
    """ 
    This block collects the history of a quantity for a given
    interval length, and it outputs a list of values when the 
    buffer is full. Then it resets the buffer.
    
    See also :ref:`block:historyt` and :ref:`block:last_n_samples`.
    """

    Block.alias("time_slice")

    Block.config("interval", "Length of the interval to record.", default=10)

    Block.input("values", "Any signal.")

    Block.output("grouped", "List of the values in a given interval.")

    def init(self):
        self.x = []
        self.t = []

    def update(self):
        sample = self.get_input(0)
        timestamp = self.get_input_timestamp(0)

        self.x.append(sample)
        self.t.append(timestamp)

        delta = self.t[-1] - self.t[0]
        #        self.info('Delta: %.10f sec  n = %6d' % (delta, len(self.t)))
        if np.abs(delta) >= self.config.interval:
            self.set_output(0, value=self.x, timestamp=self.t[0])
            self.x = []
            self.t = []



from .border import image_border
from .compose import place_at
from .filters import torgb
from contracts import contract
from procgraph import Block, BadConfig
from procgraph.block_utils import input_check_convertible_to_rgb
from . import np

__all__ = ["ImageGrid", "make_images_grid"]


class ImageGrid(Block):
    """
        A block that creates a larger image by arranging them in a grid.

        The output is rgb, uint8.

        Inputs are passed through the "torgb" function.

    """

    Block.alias("grid")

    Block.config("cols", "Columns in the grid.", default=None)
    Block.config("bgcolor", "Background color.", default=[0, 0, 0])

    Block.config("pad", "Padding for each cell", default=0)
    Block.input_is_variable("Images to arrange in a grid.", min=1)
    Block.output("grid", "Images arranged in a grid.")

    def update(self):
        if not self.all_input_signals_ready():
            return

        n = self.num_input_signals()
        for i in range(n):
            input_check_convertible_to_rgb(self, i)

        cols = self.config.cols

        if cols is not None and not isinstance(cols, int):
            raise BadConfig("Expected an integer.", self, "cols")

        images = [self.get_input(i) for i in range(n)]

        images = list(map(torgb, images))
        canvas = make_images_grid(
            images, cols=self.config.cols, pad=self.config.pad, bgcolor=self.config.bgcolor
        )

        self.set_output(0, canvas)


@contract(images="list[>=1](array)")
def make_images_grid(images, cols=None, pad=0, bgcolor=[1, 1, 1]):
    n = len(images)
    if cols is None:
        cols = int(np.ceil(np.sqrt(n)))

    rows = int(np.ceil(n * 1.0 / cols))

    assert cols > 0 and rows > 0
    assert n <= cols * rows

    # find width and height for the grid
    col_width = np.zeros(cols, dtype="int32")
    row_height = np.zeros(rows, dtype="int32")
    for i in range(n):
        image = images[i]
        col = i % cols
        row = int((i - i % cols) / cols)
        assert 0 <= col < cols
        assert 0 <= row < rows

        if pad > 0:
            image = image_border(image, left=pad, right=pad, top=pad, bottom=pad, color=bgcolor)
        width = image.shape[1]
        height = image.shape[0]

        col_width[col] = max(width, col_width[col])
        row_height[row] = max(height, row_height[row])

    canvas_width = sum(col_width)
    canvas_height = sum(row_height)

    # find position for each col and row
    col_x = np.zeros(cols, dtype="int32")
    for col in range(1, cols):
        col_x[col] = col_x[col - 1] + col_width[col - 1]

    assert canvas_width == col_x[-1] + col_width[-1]

    row_y = np.zeros(rows, dtype="int32")
    for row in range(1, rows):
        row_y[row] = row_y[row - 1] + row_height[row - 1]
    assert canvas_height == row_y[-1] + row_height[-1]

    canvas = np.zeros((canvas_height, canvas_width, 3), dtype="uint8")
    for k in range(3):
        canvas[:, :, k] = bgcolor[k] * 255

    for i in range(n):
        col = i % cols
        row = int((i - i % cols) / cols)
        assert 0 <= col < cols
        assert 0 <= row < rows
        image = images[i]
        x = col_x[col]
        y = row_y[row]

        # Pad if not right shape
        extra_hor = col_width[col] - image.shape[1]
        extra_ver = row_height[row] - image.shape[0]
        eleft = int(extra_hor / 2)
        eright = extra_hor - eleft
        etop = int(extra_ver / 2)
        ebottom = extra_ver - etop
        image = image_border(image, left=eleft, right=eright, top=etop, bottom=ebottom, color=bgcolor)

        # TODO: align here
        place_at(canvas, image, x, y)

    return canvas



from procgraph import Block

__all__ = ["Gain"]


class Gain(Block):
    """ A simple example of a gain block. """

    Block.alias("gain")

    Block.config("k", "Multiplicative gain")

    Block.input("in", "Input value")
    Block.output("out", "Output multiplied by k.")

    def update(self):
        self.output[0] = self.input[0] * self.config.k



import traceback
from copy import deepcopy
import os
import re
import sys

from pyparsing import ParseResults

from .block_config import resolve_config
from .block_meta import VARIABLE, DEFINED_AT_RUNTIME
from .constants import STRICT
from .exceptions import BadMethodCall, PGException, SemanticError, x_not_found, aslist
from .model import Model
from .model_io import ModelInput
from .parsing_elements import ParsedSignalList, VariableReference, ParsedBlock, ParsedModel, ParsedSignal
from .registrar import default_library
from .visualization import debug as debug_main, semantic_warning, info


def check_link_compatibility_input(previous_block, previous_link):
    assert isinstance(previous_link, ParsedSignalList)

    num_required = len(previous_link.signals)
    num_found = previous_block.num_output_signals()

    if num_required > num_found:
        msg = "Required at least %d, found only %d." % (num_required, num_found)
        raise SemanticError(msg, previous_link)
    # XXX, still something not quite right

    # We check that we have good matches for the previous
    for i, s in enumerate(previous_link.signals):
        assert isinstance(s, ParsedSignal)

        if s.block_name is not None:
            msg = "Could not give the block a name when the connection is " "between two blocks."
            raise SemanticError(msg, s)

        if s.local_input is None:
            s.local_input = i

        if not previous_block.is_valid_output_name(s.local_input):
            msg = 'Could not find output name "%s"(%s) in %s' % (
                s.local_input,
                type(s.local_input),
                previous_block,
            )
            raise SemanticError(msg, s)

        s.local_input = previous_block.canonicalize_output(s.local_input)


def check_link_compatibility_output(block, previous_link):
    assert isinstance(previous_link, ParsedSignalList)

    # we check that we have good matches for the next
    for i, s in enumerate(previous_link.signals):
        assert isinstance(s, ParsedSignal)

        if s.local_output is None:
            s.local_output = i

        if not block.is_valid_input_name(s.local_output):
            msg = "Could not find input name %r in %s" % (s.local_output, block)
            raise SemanticError(msg, s)

        s.local_output = block.canonicalize_input(s.local_output)


def expand_references_in_string(s, function):
    """ Expands references of the kind ${var} in the string s.
        ``function``(var) translates from var -> value """
    while True:
        m = re.match("(.*)\$\{(\w+)\}(.*)", s)
        if not m:
            return s
        before = m.group(1)
        var = m.group(2)
        after = m.group(3)
        sub = function(var)
        s = before + str(sub) + after


def create_from_parsing_results(parsed_model, name=None, config={}, library=None):
    assert isinstance(parsed_model, ParsedModel)

    def debug(s):
        if False:
            debug_main("Creating %s:%s | %s" % (name, parsed_model.name, s))

    # debug_main('config: %s' % config)

    if library is None:
        library = default_library

    model = Model(name=name, model_name=parsed_model.name)
    model.define_input_signals_new([x.name for x in parsed_model.input])
    model.define_output_signals_new([x.name for x in parsed_model.output])

    # first we divide the config in normal, and recursive config
    normal_config = {}
    recursive_config = {}
    for key, value in list(config.items()):
        if "." in key:
            recursive_config[key] = value
        else:
            normal_config[key] = value

    # We mix the normal config with the defaults
    # FIXME: None -> no where information
    #        In fact, parsed_model is the one we are instancing, NOT
    #        where the error is made.
    resolved = resolve_config(parsed_model.config, normal_config, None)
    # we give none so that it can be filled in by the caller

    # Remember config statement
    key2element = {}
    for c in parsed_model.config:
        key2element[c.variable] = c

    # We collect here all the properties, to use in initialization.
    all_config = []  # tuple (key, value, parsing_element)

    # 1. We put the resolved configuration
    for key, value in list(resolved.items()):
        all_config.append((key, value, key2element.get(key, None)))

    # 2. We also put the recursive conf
    for key, value in list(recursive_config.items()):
        all_config.append((key, value, key2element.get(key, None)))

    # 3. We process the assignments
    for assignment in parsed_model.assignments:
        # We make sure we are not overwriting configuration
        if assignment.key in resolved:
            msg = (
                "Assignment to %r overwrites a config variable. "
                "Perhaps you want to change the default instead?" % assignment.key
            )
            raise SemanticError(msg, assignment)

        all_config.append((assignment.key, assignment.value, assignment))

    # Next, define the properties hash, and populate it intelligentily
    # from the tuples in all_config.
    properties = {}
    # We keep track of what properties we use
    used_properties = set()  # of strings
    # This instead collects the block_properties
    block_properties = {}  # {str: {str: *}}

    def expand_value(value, element=None):
        """ Function that looks for VariableReference and does the
            substitution.
        """
        if isinstance(value, VariableReference):
            variable = value.variable
            if variable in os.environ:
                return os.environ[variable]
            if not variable in properties:
                msg = x_not_found("variable", variable, properties)
                raise SemanticError(msg, element)
            used_properties.add(variable)
            return expand_value(properties[variable], element=element)

        elif isinstance(value, str):
            return expand_references_in_string(
                value, lambda s: expand_value(VariableReference(s), element=element)
            )

        elif isinstance(value, dict):
            h = {}
            for key in value:
                h[key] = expand_value(value[key], element=element)
            return h

        # XXX: we shouldn't have here ParseResults
        elif isinstance(value, list) or isinstance(value, ParseResults):
            return [expand_value(s, element) for s in value]
        else:
            return value

    # We keep track of the blocks we reference so we can check it later
    referenced_blocks = []  # list of tuples (block name, parsed_element)
    for key, value, element in all_config:
        # if it is of the form  object.property = value
        if "." in key:
            # TODO: put this in syntax
            object_, property_ = key.split(".", 1)
            if not object_ in block_properties:
                block_properties[object_] = {}
            # else:
            # # XXX probably should be better
            # if not isinstance(properties[object], dict):
            #     msg = ('Error while processing "%s=%s": I already know'
            #            ' the key.' % (key, value))
            #     raise SemanticError(msg, element)
            referenced_blocks.append((object_, element))
            block_properties[object_][property_] = value  # XX or expand?
        else:
            properties[key] = expand_value(value, element=element)

    # Make sure we can access python modules in the same directory as the
    # filename; we add the directory to the sys.path

    old_sys_path = list(sys.path)
    if parsed_model.where.filename is not None:
        dirname = os.path.dirname(parsed_model.where.filename)
        if dirname is not None:
            sys.path.append(dirname)

    try:
        for x in parsed_model.imports:
            package = x.package
            if not package in sys.modules:
                info("Importing package %r..." % package)

                try:
                    __import__(package)
                except Exception as e:
                    msg = "Could not import package %r: %s" % (package, e)
                    raise SemanticError(msg, x)
    finally:
        sys.path = old_sys_path

    # Then we instantiate all the blocks

    # Iterate over connections
    for connection in parsed_model.connections:
        previous_block = None
        previous_link = None

        # print "Looking at connection %s" % connection.elements
        for i, element in enumerate(connection.elements):
            if isinstance(element, ParsedSignalList):
                # We make a copy, because later we modify (for convenience)
                # its fields, in the check_compatibility_* functions
                # For example, we assign names to input/output.
                element = deepcopy(element)

                # if this is not the last one, just save it, it will be
                # processed together with the next block
                if i != len(connection.elements) - 1:
                    previous_link = element
                else:
                    assert previous_block is not None
                    # This is the last one, we should process it with the
                    # previous_block
                    previous_link = element
                    check_link_compatibility_input(previous_block, previous_link)

                    for s in previous_link.signals:
                        # We cannot have a local output
                        if s.local_output is not None:
                            msg = "Terminator connection %s cannot have a " "local output" % s
                            raise SemanticError(msg, s)

                        if s.name in model.public_signal_names():
                            msg = "Public signal name %r already taken." % s.name
                            raise SemanticError(msg, s)

                        model.connect(
                            block1=previous_block,
                            block1_signal=s.local_input,
                            block2=None,
                            block2_signal=None,
                            public_name=s.name,
                        )

            if isinstance(element, ParsedBlock):

                # before processing this block, let's create a phantom
                # signal list
                if previous_block is not None and previous_link is None:
                    previous_link = fill_anonymous_link(previous_block)

                # also we can check right now a common error
                if previous_block is not None and previous_block.num_output_signals() == 0:
                    msg = "This block does not define outputs; yet it is not " "the last in the sequence."
                    raise SemanticError(msg, previous_block)

                block_type = expand_value(element.operation, element=element)

                # give a name if anonymous
                if element.name is None:
                    # give it, if possible the name of its type
                    if not block_type in model.name2block:
                        element.name = block_type
                    else:
                        i = 2
                        while True:
                            element.name = "%s%d" % (block_type, i)
                            i += 1
                            if not element.name in model.name2block:
                                break

                # update the configuration if given
                block_config = {}
                block_config.update(element.config)
                if element.name in block_properties:
                    more_config_for_block = block_properties[element.name]
                    # For example:
                    #   wait = 10       ->  { wait: 10 }
                    #   wait.time  = 3  ->  { wait: {time: 3} }
                    block_config.update(more_config_for_block)

                for key, value in list(block_config.items()):
                    block_config[key] = expand_value(value, element=element)

                if not library.exists(block_type):
                    msg = x_not_found("block type", block_type, library.get_known_blocks())
                    raise SemanticError(msg, element)

                debug("instancing %s:%s config: %s" % (element.name, element.operation, block_config))

                try:
                    block = library.instance(block_type=block_type, name=element.name, config=block_config)
                    block.where = element.where
                except SemanticError as e:
                    # For config (see FIXME)
                    if e.element is None:
                        e.element = element
                    raise

                # now define input and output
                generator = library.get_generator_for_block_type(block_type)

                define_input_signals(generator.input, block, previous_link, previous_block, model)

                define_output_signals(generator.output, block)

                # at this point input/output should be defined
                assert block.are_input_signals_defined()
                assert block.are_output_signals_defined()

                block = model.add_block(name=element.name, block=block)

                previous_link = None
                previous_block = block
            # end if

    # Check if any of the config referenced a nonexistent block
    # (before we warn it as just an unused variable in the next paragraph)
    for block_name, element in referenced_blocks:
        if not block_name in model.name2block:
            msg = x_not_found("block", block_name, model.name2block)
            raise SemanticError(msg, element)

    unused_properties = set(properties.keys()).difference(used_properties)
    if unused_properties:
        msg = "Unused properties: %s. (Used: %s.)" % (aslist(unused_properties), aslist(used_properties))
        if STRICT:
            raise SemanticError(msg, element=parsed_model)
        else:
            semantic_warning(msg, parsed_model)

    # One last thing: define dummy blocks for inputs without |input| blocks
    for signal in model.get_input_signals_names():
        if not signal in model.model_input_ports:
            block_type = "input"
            block_name = "dummy_input_%s" % signal
            block_config = {"name": signal}
            dummy_block = library.instance(block_type, block_name, block_config)
            generator = library.get_generator_for_block_type(block_type)
            define_input_signals(generator.input, dummy_block, None, None, model)
            define_output_signals(generator.output, dummy_block)
            model.add_block(block_name, dummy_block)
            # TODO: warn

    # TODO: warn if no output block was defined

    return model


def define_output_signals(output, block):
    # this is a special case, in which the signal name
    # is not known before parsing the configuration
    if isinstance(block, ModelInput):
        block.define_output_signals_new([block.config.name])
        return

    output_is_defined_at_runtime = len(output) == 1 and output[0].type == DEFINED_AT_RUNTIME

    if output_is_defined_at_runtime:
        try:
            names = block.get_output_signals()
        except PGException:
            raise
        except BaseException:
            raise BadMethodCall("get_output_signals", block, traceback.format_exc())

        if len(set(names)) != len(names):
            msg = "Repeated signal names in %s." % names
            raise SemanticError(msg, block)

        block.define_output_signals_new(names)
        return

    output_is_variable = len(output) == 1 and output[0].type == VARIABLE

    if output_is_variable:
        # define output signals with the same name as the input signals

        try:
            names = block.get_input_signals_names()
        except PGException:
            raise
        except BaseException:
            raise BadMethodCall("get_input_signals_names", block, traceback.format_exc())

        # TODO: maybe add a suffix someday
        # names = [name + suffix for name in names]

        block.define_output_signals_new(names)

    else:
        # simply define the output signals
        names = [x.name for x in output]
        block.define_output_signals_new(names)


def define_input_signals(input, block, previous_link, previous_block, model):  # @ReservedAssignment
    # there are two cases: either we define named signals,
    # or we have a generic number of signals
    input_is_arbitrary = len(input) == 1 and input[0].type == VARIABLE

    if input_is_arbitrary:
        # in this case, we have a minimum and maximum
        # number of signals that we can accept
        min_expected = input[0].min
        max_expected = input[0].max

        if not min_expected:
            min_expected = 0
        if not max_expected:
            max_expected = 10000

        # if we don't have a previous block, then
        # we just define no input signals
        # (if we expect something, then we throw an error)
        if previous_link is None:
            if min_expected > 0:
                msg = (
                    "I expected at least %d input signal(s) but the block "
                    "is not connected to anything." % min_expected
                )
                raise SemanticError(msg, block)
            else:
                # no inputs for this block
                block.define_input_signals_new([])
        else:
            # We have a previous link, we check that the number
            # of signals is compatible.
            num_given = len(previous_link.signals)
            ok = min_expected <= num_given <= max_expected
            if not ok:
                msg = "I expected between %d and %d input signals, and I got " "%d." % (
                    min_expected,
                    max_expected,
                    num_given,
                )
                raise SemanticError(msg, block)
            # Define input signals given the names
            names = []
            for i in range(num_given):
                # --> [local_input] name [local_output] -->
                if previous_link.signals[i].local_output is not None:
                    name = previous_link.signals[i].local_output
                elif previous_link.signals[i].name is not None:
                    name = previous_link.signals[i].name
                elif previous_link.signals[i].local_input is not None:
                    name = previous_link.signals[i].local_input
                else:
                    assert False

                # just in case the thing is repeated
                # TODO: warn?  if name in names: ...
                name = find_unique_name(name, names)
                names.append(name)
            block.define_input_signals_new(names)
    else:  # the input is not arbitrary
        # define right away the names, it does not depend
        # on anything else
        names = [x.name for x in input]
        block.define_input_signals_new(names)

        # now check we were given the right input
        num_expected = len(names)

        # if we expect something and it is not given,
        # raise an exception
        if previous_link is None:
            if num_expected > 0:
                msg = "The block expected at least %d input signals" " but none were given." % num_expected
                raise SemanticError(msg, block)
        else:
            # we have a previous block, the number of signals
            # should match
            num_given = len(previous_link.signals)

            if num_expected != num_given:
                msg = "The block expected %d input signals, got %d." % (num_expected, num_given)
                raise SemanticError(msg, block)

    # print "Defined block %s = %s " % (element.name , block)
    if previous_link is not None:
        check_link_compatibility_output(block, previous_link)

        if previous_block is not None:
            # normal connection between two blocks with named signals

            # Here we have to make sure that, if the blocks defined
            #  signals input/outputs, then the signals given by the user
            #  are coherent.

            check_link_compatibility_input(previous_block, previous_link)

            # Finally we create the connection
            for s in previous_link.signals:
                if s.name is None:
                    s.name = "input_%s_for_%s" % (s.local_output, block)
                s_name = find_unique_name(s.name, model.public_signal_names())
                model.connect(previous_block, s.local_input, block, s.local_output, s_name)
        else:
            # this is the first block with previous signals
            # this time we need to be careful, because
            # links can refer to other parts
            for s in previous_link.signals:
                # Cannot use local_input here
                if s.local_input is not None:
                    msg = "Link %s uses local input without antecedent. " % s
                    raise SemanticError(msg, s)
                # Check if it is using an explicit block name
                if s.block_name is not None:
                    if not s.block_name in model.name2block:
                        msg = "Link %s refers to an unknown block %r." "Valid blocks: %s." % (
                            s,
                            s.block_name,
                            aslist(model.name2block),
                        )
                        raise SemanticError(msg, s)

                    input_block = model.name2block[s.block_name]
                    if not input_block.is_valid_output_name(s.name):
                        # TODO: make other friendly messages like this
                        msg = "This link refers to an unknown output %r. " % s.name
                        msg += "The known outputs are: %s." % input_block.get_output_signals_names()
                        # msg += "  link: %s \n" % s
                        # msg += " block: %s \n" % input_block
                        raise SemanticError(msg, element=s)
                    s.local_input = input_block.canonicalize_output(s.name)
                else:
                    if not s.name in model.name2block_connection:
                        # throw out the autogenerated (XXX: make a flag)
                        valid = [x for x in model.name2block_connection if not ":" in x]
                        msg = "This link refers to an unknown signal %r.\n" "Valid signals: %s." % (
                            s.name,
                            aslist(valid),
                        )
                        raise SemanticError(msg, element=s)
                    defined_signal = model.name2block_connection[s.name]
                    input_block = defined_signal.block1
                    s.local_input = defined_signal.block1_signal

                # make up a unique name
                name = "%s:%s:%s" % (input_block.name, s.local_output, block.name)
                name = find_unique_name(name, model.public_signal_names())

                model.connect(input_block, s.local_input, block, s.local_output, name)


def find_unique_name(prefix, existing):
    name = prefix
    count = 1
    while name in existing:
        count += 1
        name = "%s%d" % (prefix, count)
    return name


def fill_anonymous_link(previous_block):
    # --> [local_input] name [local_output] -->
    names = previous_block.get_output_signals_names()
    signals = []
    for name in names:
        signal = ParsedSignal(name=None, block_name=None, local_input=name, local_output=None)
        signals.append(signal)
    return ParsedSignalList(signals)



from procgraph import Block


class Print(Block):
    """
        Print a representation of the input values along with
        their timestamp.
    """

    Block.alias("print")

    Block.input_is_variable("Signals to print.", min=1)

    def update(self):
        for i in range(self.num_input_signals()):
            print(
                ("P %s %s %s" % (self.canonicalize_input(i), self.get_input_timestamp(i), self.get_input(i)))
            )



from contracts import ContractsMeta

from .constants import FIXED, VARIABLE, DEFINED_AT_RUNTIME
from .exceptions import BlockWriterError, ModelWriterError


class BlockConfig(object):
    def __init__(self, variable, has_default, default, desc, desc_rest, dtype, where=None):
        self.variable = variable
        self.has_default = has_default
        self.default = default
        self.desc = desc
        self.desc_rest = desc_rest
        self.where = where
        self.dtype = dtype

    def __repr__(self):
        return "BlockConfig(%r,%r,%r,%r,%r)" % (
            self.variable,
            self.has_default,
            self.default,
            self.desc,
            self.desc_rest,
        )


class BlockInput(object):
    def __init__(self, type, name, min, max, desc, desc_rest, where, dtype):  # @ReservedAssignment
        self.type = type  # FIXED, VARIABLE, DEFINED_AT_RUNTIME
        self.name = name
        self.min = min
        self.max = max
        self.desc = desc
        self.desc_rest = desc_rest
        self.where = where
        self.dtype = dtype

    def __repr__(self):
        return "BlockInpput(%s,%s)" % (self.type, self.name)


class BlockOutput(object):
    def __init__(self, type, name, desc, desc_rest, where, dtype):  # @ReservedAssignment
        self.type = type  # FIXED, VARIABLE, DEFINED_AT_RUNTIME
        self.name = name
        self.desc = desc
        self.desc_rest = desc_rest
        self.where = where
        self.dtype = dtype

    def __repr__(self):
        return "BlockOutput(%s,%s)" % (self.type, self.name)


def block_alias(name):
    assert isinstance(name, str)
    BlockMeta.aliases.append(name)


def block_config(name, description=None, default="not-given", dtype=None):
    assert isinstance(name, str)
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    has_default = default != "not-given"
    if [x for x in BlockMeta.tmp_config if x.variable == name]:
        cleanup()
        msg = "Already described config variable %r." % name

        raise BlockWriterError(msg)
    BlockMeta.tmp_config.append(
        BlockConfig(name, has_default, default, desc, desc_rest, dtype=dtype, where=None)
    )


def block_input(name, description=None, dtype=None):
    assert isinstance(name, str)
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    if [x for x in BlockMeta.tmp_input if x.name == name]:
        cleanup()
        msg = 'Already described input variable "%s".' % name
        raise BlockWriterError(msg)
    if BlockMeta.tmp_input and BlockMeta.tmp_input[-1].type == VARIABLE:
        cleanup()
        msg = "Cannot mix variable and fixed input."
        raise BlockWriterError(msg)
    BlockMeta.tmp_input.append(BlockInput(FIXED, name, None, None, desc, desc_rest, None, dtype=None))


def block_input_is_variable(description=None, min=None, max=None):  # @ReservedAssignment
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    if BlockMeta.tmp_input:
        cleanup()
        msg = "Cannot mix variable and fixed input or variable with variable."
        raise BlockWriterError(msg)
    BlockMeta.tmp_input.append(BlockInput(VARIABLE, None, min, max, desc, desc_rest, None, dtype=None))


def block_output(name, description=None, dtype=None):
    assert isinstance(name, str)
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    if [x for x in BlockMeta.tmp_output if x.name == name]:
        cleanup()
        msg = "Already described output variable %r." % name
        raise BlockWriterError(msg)
    if BlockMeta.tmp_output and BlockMeta.tmp_output[-1].type == VARIABLE:
        cleanup()
        msg = "Cannot mix variable and fixed output."
        raise BlockWriterError(msg)

    BlockMeta.tmp_output.append(BlockOutput(FIXED, name, desc, desc_rest, None, dtype=dtype))


def block_output_is_variable(description=None, suffix=None):
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    if BlockMeta.tmp_output:
        cleanup()
        msg = "Cannot mix variable and fixed output or variable with variable. " "(added already: %s)" % (
            BlockMeta.tmp_output
        )
        raise BlockWriterError(msg)
    bo = BlockOutput(VARIABLE, suffix, desc, desc_rest, None, dtype=None)
    BlockMeta.tmp_output.append(bo)


def block_output_is_defined_at_runtime(description=None):
    assert description is None or isinstance(description, str)
    desc, desc_rest = split_docstring(description)
    if BlockMeta.tmp_output:
        cleanup()
        raise BlockWriterError(
            "Cannot mix variable and fixed output"
            " or variable with variable. (added already: %s)" % (BlockMeta.tmp_output)
        )
    BlockMeta.tmp_output.append(BlockOutput(DEFINED_AT_RUNTIME, None, desc, desc_rest, None, dtype=None))


def cleanup():
    """ Cleans up temporary data for the meta-sugar, if the construction 
        is aborted """
    BlockMeta.tmp_config = []
    BlockMeta.tmp_output = []
    BlockMeta.tmp_input = []
    BlockMeta.aliases = []


class BlockMeta(ContractsMeta):
    aliases = []
    tmp_config = []
    tmp_input = []
    tmp_output = []

    def __init__(cls, clsname, bases, clsdict):  # @UnusedVariable @NoSelf
        ContractsMeta.__init__(cls, clsname, bases, clsdict)
        # Do not do this for the superclasses
        if clsname in ["Generator", "Block"]:
            return

        setattr(cls, "defined_in", cls.__module__)

        setattr(cls, "config", BlockMeta.tmp_config)
        setattr(cls, "output", BlockMeta.tmp_output)
        setattr(cls, "input", BlockMeta.tmp_input)
        BlockMeta.tmp_config = []
        BlockMeta.tmp_output = []
        BlockMeta.tmp_input = []

        has_variable_input = [x for x in BlockMeta.tmp_input if x.type == VARIABLE]
        has_variable_output = [x for x in BlockMeta.tmp_output if x.type == VARIABLE]

        if has_variable_output and not has_variable_input:
            raise ModelWriterError("Cannot have variable output without " "variable input.")

        if len(BlockMeta.aliases) > 1:
            raise ModelWriterError(
                "We don't support multiple aliases yet. " "Tried to set  %r." % BlockMeta.aliases
            )

        if BlockMeta.aliases:
            name = BlockMeta.aliases[0]
        else:
            name = cls.__name__

        from .registrar import default_library

        default_library.register(name, cls)

        BlockMeta.aliases = []


class BlockMetaSugar(object):
    @staticmethod
    def alias(*arg, **kwargs):
        block_alias(*arg, **kwargs)

    @staticmethod
    def config(*arg, **kwargs):
        block_config(*arg, **kwargs)

    @staticmethod
    def input(*arg, **kwargs):  # @ReservedAssignment
        block_input(*arg, **kwargs)

    @staticmethod
    def output(*arg, **kwargs):
        block_output(*arg, **kwargs)

    @staticmethod
    def output_is_variable(*arg, **kwargs):
        block_output_is_variable(*arg, **kwargs)

    @staticmethod
    def output_is_defined_at_runtime(description=None):
        """ If this is used, then one must implement get_output_signals() """
        block_output_is_defined_at_runtime(description)

    @staticmethod
    def input_is_variable(description=None, min=None, max=None):  # @ReservedAssignment
        """ Declares that this block can accept a variable number
            of inputs. You can specify minimum and maximum number. """
        block_input_is_variable(description, min, max)


# TODO: move this somewhere else


def trim(docstring):
    if not docstring:
        return ""
    lines = docstring.expandtabs().splitlines()

    # Determine minimum indentation (first line doesn't count):
    maxi = 10000
    indent = maxi
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))

    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < maxi:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())

    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    result = "\n".join(trimmed)

    # print 'input: """%s"""' % docstring
    # print 'result: """%s"""' % result
    return result


# TODO: remove space on the right


def split_docstring(s):
    """ Splits a docstring in a tuple (first, rest). """
    if s is None:
        return None, None
    s = trim(s)
    all_lines = s.split("\n")
    stripped = [l.strip() for l in all_lines]
    valid_lines = [l for l in stripped if l]
    if valid_lines:
        for i in range(len(all_lines)):
            if all_lines[i]:  # found first
                # join all non-empty lines with the first
                j = i
                while j < len(all_lines) - 1 and all_lines[j].strip():
                    j += 1
                first = " ".join(all_lines[i : (j + 1)])
                rest = "\n".join(all_lines[j + 1 :])
                return first, rest
        assert False
    else:
        return None, None



""" 
    Blocks for encoding/decoding video based on MPlayer. 

    Needs ffmepg > 0.8.6 that is present in Ubuntu 12.04. No PPA available. 
    Download static version from http://ffmpeg.gusari.org/
    (http://ffmpeg.gusari.org/static/64bit/ffmpeg.static.64bit.latest.tar.gz)
    and place in path.
"""


from .depth_buffer import *
from .fix_frame_rate import *
from .mencoder import *
from .mp4conversion import convert_to_mp4
from .mplayer import *
from .quick_animation import *


procgraph_info = {
    # List of python packages
    "requires": ["qtfaststart"]
}


from procgraph import pg_add_this_package_models

pg_add_this_package_models(__file__, __package__)



#!/usr/bin/env python
from contracts import contract
from itertools import product
from optparse import OptionParser
from procgraph_images import TransparentImageAverage
from procgraph_pil import imread, imwrite
from reprep import scale
import numpy as np
import os
from procgraph.utils.calling_ext_program import system_cmd_result
from procgraph.utils.friendly_paths import friendly_path


usage = """
    %cmd  [-o <output>] <filename>
"""


def main():
    parser = OptionParser(usage=usage)

    parser.add_option("-i", "--input", dest="video", type="string", help="Input video.")
    parser.add_option("-o", "--output", dest="background", type="string", help="Output png.")
    parser.add_option(
        "-f",
        "--frames",
        dest="frames",
        type="string",
        default="1.0,2,3.4",
        help="Frames to use; floats separated by comma.",
    )

    (options, args) = parser.parse_args()

    if args:
        msg = "No argument necessary."
        raise Exception(msg)

    video = options.video
    background = options.background
    frames = list(map(float, options.frames.split(",")))

    result = find_background(video, frames, debug=True)
    imwrite(result["background"], background)
    return 0


def get_frame(video, when):
    # out = join(tmp_path, 'frame%s.png' % when)
    out = os.path.basename(video) + ".tmp_frame%f.png" % when
    cmd = "ffmpeg -ss %s -i %s -y -f image2 -vframes 1 %s" % (when, video, out)
    system_cmd_result(
        cwd=".",
        cmd=cmd,
        display_stdout=False,
        display_stderr=False,
        raise_on_error=True,
        capture_keyboard_interrupt=False,
    )  # @UnusedVariable

    frame = imread(out)
    os.unlink(out)
    return frame


from os.path import dirname, exists, basename, splitext, join


@contract(whens="list[>=3](float)")
def find_background(video, whens, debug=False, tmp_path=None):
    if not os.path.exists(video):
        msg = "Filename does not exist: %s" % friendly_path(video)
        raise ValueError(msg)
    id_video = splitext(basename(video))[0]

    if tmp_path is None:
        tmp_path = join(dirname(video), "find_background-%s" % id_video)
    if not exists(tmp_path):
        os.makedirs(tmp_path)

    @contract(rgb="array[HxWx3](uint8)")
    def write_debug(rgb, id_image):
        if debug:
            f = join(tmp_path, "%s.png" % id_image)
            # print('Writing %s' % f)
            imwrite(rgb, f)

    frames = []
    for i, when in enumerate(whens):
        frame = get_frame(video, when)
        write_debug(frame, "frame-%s-t%4.2f" % (i, when))
        frames.append(frame.astype("float32"))

    n = len(frames)

    errs = {}
    for i, j in product(list(range(n)), list(range(n))):
        if i == j:
            continue

        diff = np.abs(frames[i] - frames[j])
        errs[(i, j)] = np.mean(diff, axis=2)

        write_debug(scale(errs[(i, j)]), "errs-%s-%s" % (i, j))

    mm = np.dstack(tuple(errs.values()))
    s = np.mean(mm, axis=2)
    write_debug(scale(s), "s")

    # print('computing moving stuff')
    moving = []
    bgshade = []
    for i in range(n):
        its = np.dstack([errs[(i, j)] for j in range(n) if i != j])
        moving_i = np.min(its, axis=2)
        moving.append(moving_i)

        bgshade_i = assemble_transp(frames[i], 1.0 / (moving_i + 10))
        bgshade.append(bgshade_i)

        write_debug(scale(moving_i), "moving-%d" % i)
        write_debug(bgshade_i, "bgshade-%d" % i)

    avg = TransparentImageAverage()
    for i in range(n):
        avg.update(bgshade[i])
    bg_rgb = avg.get_rgb()

    write_debug(bg_rgb, "background")

    return {"background": bg_rgb}


@contract(rgb="array[HxWx3]", alpha="array[HxW]", returns="array[HxWx4](uint8)")
def assemble_transp(rgb, alpha):
    H, W = rgb.shape[0], rgb.shape[1]
    rgba = np.zeros((H, W, 4), dtype="uint8")
    rgba[:, :, 0:3] = rgb[:, :, 0:3]
    # print('alpha range %f - %f' % (np.min(alpha), np.max(alpha)))
    alpha = alpha - np.min(alpha)
    alpha = alpha / np.max(alpha)
    alpha = alpha * 255
    rgba[:, :, 3] = alpha[:, :]
    return rgba


if __name__ == "__main__":
    main()



from .utils import PGTestCase

good_examples = [
    "import a",
    "import a.b",
    "import a.b.c",
    "import a .b. c",
    "import a. b .c",
    "import a .b. c",
]

bad_examples = ["import 0a", "import 0a.", "import .0a"]


class SyntaxTestImport(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples:
            self.check_syntax_error(example)

    def testExamples(self):
        for example in good_examples:
            self.check_syntax_ok(example)



from system_cmd import *  # @UnusedWildImport

#
# from procgraph.utils import indent
# import subprocess
# import tempfile
# import sys
# import os
#
# __all__ = ['system_cmd_result', 'CmdResult', 'CmdException']
#
#
# class CmdResult(object):
#     def __init__(self, cwd, cmd, ret, rets, interrupted, stdout, stderr):
#         self.cwd = cwd
#         self.cmd = cmd
#         self.ret = ret
#         self.rets = rets
#         self.stdout = stdout
#         self.stderr = stderr
#         self.interrupted = interrupted
#
#     def __str__(self):
#         msg = ('The command: %s\n'
#                '     in dir: %s\n' % (self.cmd, self.cwd))
#
#         if self.interrupted:
#             msg += 'Was interrupted by the user\n'
#         else:
#             msg += 'returned: %s' % self.ret
#         if self.rets is not None:
#             msg += '\n' + indent(self.rets, 'error>')
#         if self.stdout:
#             msg += '\n' + indent(self.stdout, 'stdout>')
#         if self.stderr:
#             msg += '\n' + indent(self.stderr, 'stderr>')
#         return msg
#
#
# class CmdException(Exception):
#     def __init__(self, cmd_result):
#         Exception.__init__(self, cmd_result)
#         self.res = cmd_result
#
#     def __str__(self):
#         return self.res.__str__()
#
#
# def system_cmd_result(cwd, cmd,
#                       display_stdout=False,
#                       display_stderr=False,
#                       raise_on_error=False,
#                       write_stdin='',
#                       capture_keyboard_interrupt=False):  # @UnusedVariable
#     '''
#         Returns the structure CmdResult; raises CmdException.
#         Also OSError are captured.
#         KeyboardInterrupt is passed through unless specified
#
#         :param write_stdin: A string to write to the process.
#     '''
#     tmp_stdout = tempfile.TemporaryFile()
#     tmp_stderr = tempfile.TemporaryFile()
#
#     ret = None
#     rets = None
#     interrupted = False
#
#     try:
#         stdout = None if display_stdout else tmp_stdout.fileno()
#         stderr = None if display_stderr else tmp_stderr.fileno()
#         p = subprocess.Popen(
#                 cmd2args(cmd),
#                 stdin=subprocess.PIPE,
#                 stdout=stdout,
#                 stderr=stderr,
#                 cwd=cwd)
#
#         if write_stdin != '':
#             p.stdin.write(write_stdin)
#             p.stdin.flush()
#         p.stdin.close()
#         p.wait()
#         ret = p.returncode
#         rets = None
#         interrupted = False
#     except KeyboardInterrupt:
#         ret = 100
#         interrupted = True
#     except OSError as e:
#         interrupted = False
#         ret = 200
#         rets = str(e)
#
#     # remember to go back
#     def read_all(f):
#         os.lseek(f.fileno(), 0, 0)
#         return f.read()
#
#     res = CmdResult(cwd, cmd, ret, rets, interrupted,
#                     stdout=read_all(tmp_stdout),
#                     stderr=read_all(tmp_stderr))
#
#     if raise_on_error:
#         if res.ret != 0:
#             raise CmdException(res)
#
#     return res
#
#
# def cmd2args(s):
#     ''' if s is a list, leave it like that; otherwise split()'''
#     if isinstance(s, list):
#         return s
#     elif isinstance(s, str):
#         return s.split()
#     else:
#         assert False
#
# if __name__ == '__main__':
#     cwd = '.'
#     cmd = " ".join(sys.argv[1:])
#     cmd = cmd.strip()
#     if not cmd:
#         print('No command given.')
#     else:
#         try:
#             display = False
#             res = system_cmd_result(cwd, cmd,
#                           display_stdout=display,
#                           display_stderr=display,
#                           raise_on_error=True)
#             print('Succeed: %s' % res)
#         except CmdException as e:
#             print('Exception: %s' % e)



"""
    These are three proxies for accessing input, output, and state.
    
    These allow to write::
    
        self.output.y = self.input.x * self.config.gain
    
    instead of::
        self.set_output('y', self.get_input('x') * self.get_config('gain'))
        
"""


class Proxy(object):
    def __init__(self, block):
        self.__dict__["block"] = block


class InputProxy(Proxy):
    def __getattr__(self, i):
        return self.__dict__["block"].get_input(i)

    def __getitem__(self, key):
        return self.__dict__["block"].get_input(key)

    def __setattr__(self, i, v):
        raise ValueError("You can only read input, not change it.")

    def __setitem__(self, key, value):
        raise ValueError("You can only read input, not change it.")


class OutputProxy(Proxy):
    def __getattr__(self, i):
        raise ValueError("You can only set the output.")

    def __getitem__(self, key):
        raise ValueError("You can only set the output.")

    def __setattr__(self, i, v):
        self.__dict__["block"].set_output(i, v)

    def __setitem__(self, key, value):
        self.__dict__["block"].set_output(key, value)


class StateProxy(Proxy):
    def __getattr__(self, i):
        return self.__dict__["block"].get_state(i)

    def __getitem__(self, key):
        return self.__dict__["block"].get_state(key)

    def __setattr__(self, i, v):
        self.__dict__["block"].set_state(i, v)

    def __setitem__(self, key, value):
        self.__dict__["block"].set_state(key, value)


class ConfigProxy(Proxy):
    def __getattr__(self, i):
        return self.__dict__["block"].get_config(i)

    def __getitem__(self, key):
        return self.__dict__["block"].get_config(key)

    def __setattr__(self, i, v):
        raise ValueError("You can only read config %r, not change it." % i)

    def __setitem__(self, key, value):
        raise ValueError("You can only read config %r, not change it." % key)



from procgraph import register_model_spec

# XXX: make models/ directory

register_model_spec(
    """
--- model variance 
''' Computes the element-wise variance. '''
input x 'Any numpy array'
output var_x 'Variance of ``x``.'
config wait = 100 "Number of samples to wait before declaring the value valid."

|input name=x| --> x --> |expectation| --> |wait n=$wait| --> Ex

   x, Ex --> |sync| --> |-| --> error 
   
   error -> |square| --> |expectation| --> |output name=var_x|
    
"""
)

register_model_spec(
    """
--- model soft_variance 
''' Computes the element-wise "soft" variance (exp. of error abs. value) '''
config wait = 100 "Number of samples to wait before declaring the value valid."
input x 'Any numpy array'
output var_x 'Soft variance of ``x``.'

|input name=x| --> x --> |expectation| --> |wait n=$wait| --> Ex

   x, Ex --> |sync| --> |-| --> error 
   
   error -> |abs| --> |expectation| --> |output name=var_x|
    
"""
)



from . import np

from procgraph import Block


# TODO: move somewhere else?
def to_rgb(rgb):
    """ Remove the alpha layer if present. """
    if rgb.shape[2] == 3:
        return rgb

    return rgb[:, :, 0:3]


# TODO: move somewhere else?
def to_rgba(rgb):
    """ Adds a opaque alpha if not present. """
    if rgb.shape[2] == 4:
        rgba = rgb.copy()
        return rgba

    rgba = np.ndarray((rgb.shape[0], rgb.shape[1], 4), dtype="uint8")
    rgba[:, :, 0] = rgb[:, :, 0]
    rgba[:, :, 1] = rgb[:, :, 1]
    rgba[:, :, 2] = rgb[:, :, 2]
    rgba[:, :, 3] = 255

    return rgba


def blend(a, b):
    """ Blends two RGB or RGBA images. """
    if (a.shape[0] != b.shape[0]) or (a.shape[1] != b.shape[1]):
        raise ValueError("Incompatible shapes: %s and %s." % (a.shape, b.shape))

    # Add alpha component if missing
    a = to_rgba(a)
    b = to_rgba(b)

    # alpha_a = a[:, :, 3].astype('float32')
    alpha_b = b[:, :, 3].astype("float32")

    # weight_a = alpha_a / (alpha_a + alpha_b)
    weight_b = alpha_b / 255.0
    weight_a = -weight_b + 1

    result = np.ndarray(shape=(a.shape[0], a.shape[1], 4), dtype="uint8")

    for i in [0, 1, 2]:
        result[:, :, i] = a[:, :, i] * weight_a + b[:, :, i] * weight_b

    result[:, :, 3] = 255

    return result.astype("uint8")


class Blend(Block):
    """
        Blend two or more images.
        
        RGB images are interpreted as having full alpha (opaque). 
        All images must have the same width and height.
        
    """

    Block.alias("blend")

    Block.input_is_variable("images to blend", min=2)

    Block.output("rgb", "The output is a RGB image (no alpha)")

    def update(self):
        # TODO: check images
        result = None
        for signal in self.get_input_signals_names():
            image = self.get_input(signal)
            if result is None:
                result = image
                continue

            result = blend(result, image)

        self.output.rgb = to_rgb(result)



from .utils import PGTestCase

good_examples = [
    """
type = generic_in0_out1

|$type | --> X
""",
    """
type = gen

|"${type}eric_in0_out1"| --> X
""",
]


bad_examples = [
    """

|"${type}eric_in0_out1"| --> X
"""
]


class DynamicBlocksTest(PGTestCase):
    def testGoodExamples(self):
        for example in good_examples:
            self.check_semantic_ok(example)

    def testBadExamples(self):
        for example in bad_examples:
            self.check_semantic_error(example)



""" 
    Blocks using Matplotlib to display data. 

"""

procgraph_info = {"requires": ["matplotlib", "matplotlib.pylab"]}

# Smart dependencies initialization
from procgraph import import_magic, import_succesful

matplotlib = import_magic(__name__, "matplotlib")
if import_succesful(matplotlib):
    if matplotlib.get_backend() != "agg":
        matplotlib.use("agg")

pylab = import_magic(__name__, "matplotlib.pylab")

from .pylab_to_image import pylab2rgb
from .plot_anim import PlotAnim
from .plot import *
from .plot_generic import *

# TODO: plot should fail for no inputs

from procgraph import pg_add_this_package_models

pg_add_this_package_models(__file__, __package__)



from ..core.exceptions import BadInput
from contracts.interface import describe_value


# TODO: make naming uniform
def check_2d_array(value, name=None):
    import numpy as np

    """ Checks that we have 2D numpy array """
    if not isinstance(value, np.ndarray):
        raise BadInput("Expected 2d array, got %s." % value.__class__.__name__, None, name)

    if len(value.shape) != 2:
        raise BadInput("Expected 2D array, got %s." % str(value.shape), None, name)


def assert_rgb_image(image, name=None):
    import numpy as np

    if not isinstance(image, np.ndarray):
        raise BadInput("Expected RGB image for %r, got %s." % (name, image.__class__.__name__), None, name)

    if image.dtype != "uint8":
        raise BadInput(
            "Expected RGB image for %r, got an array %s %s." % (name, str(image.shape), image.dtype),
            None,
            name,
        )

    if len(image.shape) != 3 or image.shape[2] != 3:
        raise BadInput("Bad shape for %s, expected RGB, got %s." % (name, str(image.shape)), None, name)


def assert_gray_image(image, name=None):
    import numpy as np

    if not isinstance(image, np.ndarray):
        raise BadInput(
            "Expected a grayscale image for %r, got %s." % (name, image.__class__.__name__), None, name
        )

    if image.dtype != "uint8":
        raise BadInput(
            "Expected a grayscale image for %r, got an array %s %s." % (name, str(image.shape), image.dtype),
            None,
            name,
        )

    if len(image.shape) != 2:
        raise BadInput("Bad shape for %r, expected grayscale, got %s." % (name, str(image.shape)), None, name)


def check_numpy_array(value):
    import numpy as np

    if not isinstance(value, np.ndarray):
        msg = "Expected array: %s" % describe_value(value)
        raise ValueError(msg)


def check_convertible_to_rgb(image):
    check_numpy_array(image)

    if image.ndim == 3:
        good = image.shape[2] in [3, 4]
        if not good:
            msg = "Wrong dimensions: %s" % image.shape
            raise ValueError(msg)

    if image.ndim > 3 or image.ndim <= 1:
        msg = "Bad shape: %s" % image.shape
        raise ValueError(msg)

    if image.dtype == "uint8":
        pass
    elif image.dtype == "float32":
        imin = image.min()
        imax = image.max()
        ok = imin >= 0 and imax <= 1
        if not ok:
            msg = "Invalid values range: %s %s" % (imin, imax)
            raise ValueError(msg)
    else:
        msg = "invalid dtype %r" % image.dtype
        raise ValueError(msg)


def input_check_numpy_array(block, signal):
    try:
        check_numpy_array(block.get_input(signal))
    except ValueError as e:
        raise BadInput(str(e), block, signal) from e


def input_check_convertible_to_rgb(block, signal):
    try:
        check_convertible_to_rgb(block.get_input(signal))
    except ValueError as e:
        raise BadInput(str(e), block, signal) from e


def input_check_rgb_or_grayscale(block, input):  # @ReservedAssignment
    """ Checks that the selected input is either a grayscale or RGB image.
        That is, a numpy array of uint8 either H x W,  H x W x 3,
        or HxWx4. 
        Raises BadInput if it is not. 
    """
    # TODO: write this better

    image = block.get_input(input)
    input_check_numpy_array(block, input)

    if image.dtype != "uint8":
        msg = "Expected an image, got an array %s %s." % (str(image.shape), image.dtype)
        raise BadInput(msg, block, input)

    shape = image.shape
    if len(shape) == 3:
        depth = shape[2]
        if not depth in [3, 4]:
            msg = "Bad shape for image: %s" % str(shape)
            raise BadInput(msg, block, input)
    elif len(shape) == 2:
        pass
    else:
        msg = "Bad shape for image: %s" % str(shape)
        raise BadInput(msg, block, input)

    if shape[0] == 0 or shape[1] == 0:
        msg = "Image with zero width or height: %s" % (str(shape))
        raise BadInput(msg, block, input)


def check_rgb(block, input):  # @ReservedAssignment
    """ Checks that the selected input is either  a RGB image.
        That is, a numpy array of uint8 of shape H x W x 3. 
        Raises BadInput if it is not. 
    """
    image = block.get_input(input)
    import numpy as np

    if not isinstance(image, np.ndarray):
        raise BadInput(
            "Expected RGB, this is not even a " "numpy array: %s" % image.__class__.__name__, block, input
        )
    if image.dtype != "uint8":
        raise BadInput(
            "Expected an image, got an array %s %s." % (str(image.shape), image.dtype), block, input
        )
    shape = image.shape
    if len(shape) == 3:
        if shape[2] != 3:
            raise BadInput("Bad shape for image: %s" % str(shape), block, input)
    else:
        raise BadInput("Bad shape for image: %s" % str(shape), block, input)



from . import np
from procgraph import simple_block


# TODO: case width = None -> sqrt
# TODO: resize 2D
@simple_block
def reshape2d(x, width=None, height=None, fill_value=np.NaN, force=False):
    """ Reshapes x into (?, width) if x is 1D.
        If width is not given, it is set to be the sqrt(size).
    
        If x is already 2D, it is left alone, unless "force" is True.
    """
    x = np.array(x)
    if not force:
        if x.ndim == 2:
            return x

    if width is not None and height is not None:
        pass
    elif width is None and height is not None:
        width = np.ceil(x.size * 1.0 / height)
    elif width is not None and height is None:
        height = np.ceil(x.size * 1.0 / width)
    elif width is None and height is None:
        width = int(np.sqrt(x.size))
        height = np.ceil(x.size * 1.0 / width)

    assert width is not None
    assert height is not None
    assert width * height >= x.size

    shape = (height, width)
    y = np.zeros(shape=shape, dtype=x.dtype)
    y.fill(fill_value)
    y.flat[0 : x.size] = x

    return y



from .debug_components import *

# TODO: make this list automatically generated


def load(x):
    __import__(x)


load("procgraph_cv")
# import procgraph_cv
load("procgraph_foo")
# import procgraph_foo
load("procgraph_hdf")
# import procgraph_hdf
load("procgraph_images")
# import procgraph_images
load("procgraph_io_misc")
# import procgraph_io_misc
load("procgraph_mpl")
# import procgraph_mpl
load("procgraph_mplayer")
# import procgraph_mplayer
load("procgraph_numpy_ops")
# import procgraph_numpy_ops
load("procgraph_pil")
# import procgraph_pil
load("procgraph_robotics")
# import procgraph_robotics
load("procgraph_signals")
# import procgraph_signals
load("procgraph_statistics")
# import procgraph_statistics



from .block import Block


class ModelInput(Block):
    """ This represents one input to the model. """

    Block.alias("input")

    Block.config("name", "Input signal name.")

    Block.output("dummy")

    def init(self):
        self.signal_name = self.config.name

    def update(self):
        pass


class ModelOutput(Block):
    """ This represents one output to the model. """

    Block.alias("output")

    Block.config("name", "Output signal name")

    Block.input("dummy")

    def init(self):
        self.signal_name = self.config.name

    def update(self):
        pass



from procgraph import Block


class MakeTuple(Block):
    """ Creates a tuple out of the input signals values. 
    
        Often used for plotting two signals as (x,y); see :ref:`block:plot`.
    """

    Block.alias("make_tuple")

    Block.input_is_variable("Signals to unite in a tuple.")

    Block.output("tuple", "Tuple containing signals.")

    def update(self):
        values = self.get_input_signals_values()
        self.output.tuple = tuple(values)



from collections import namedtuple
from optparse import OptionParser
import os
import sys

from ..core.block import Block
from ..core.block_meta import split_docstring, FIXED, VARIABLE, DEFINED_AT_RUNTIME
from ..core.import_magic import get_module_info
from ..core.model_loader import ModelSpec
from ..core.registrar import default_library


# TODO: make it possible to document simple functions
type_block = "block"
type_model = "model"
type_simple_block = "simple_block"

ModelDoc = namedtuple(
    "ModelDoc", "name source module type implementation input " "output config desc desc_rest"
)

# module : reference to actual module instance
ModuleDoc = namedtuple("ModuleDoc", "name blocks desc desc_rest module procgraph_info")


def get_module_name_with_doc(original_module_name):
    assert isinstance(original_module_name, str)

    module_name = original_module_name

    while True:
        module = __import__(module_name, fromlist=["x"])
        if module.__doc__:
            return module.__name__

        parent_name = ".".join(module.__name__.split(".")[:-1])

        # Empty!!
        if not parent_name or parent_name == "procgraph":
            return original_module_name

        module_name = parent_name


def collect_info(block_type, block_generator):

    if isinstance(block_generator, ModelSpec):
        parsed_model = block_generator.parsed_model
        type = type_model  # @ReservedAssignment
        implementation = None
        assert isinstance(block_generator.defined_in, str)
        module = get_module_name_with_doc(block_generator.defined_in)

        source = parsed_model.where.filename
        desc, desc_rest = split_docstring(parsed_model.docstring)

        config = parsed_model.config
        input = parsed_model.input  # @ReservedAssignment
        output = parsed_model.output

    elif issubclass(block_generator, Block):
        type = type_block  # @ReservedAssignment
        implementation = block_generator

        if block_generator.defined_in:
            original_module_name = block_generator.defined_in
            assert isinstance(block_generator.defined_in, str)
        else:
            original_module_name = block_generator.__module__

        assert isinstance(original_module_name, str)

        original_module = __import__(original_module_name, fromlist=["x"])

        source = original_module.__file__
        module = get_module_name_with_doc(original_module_name)

        desc, desc_rest = split_docstring(block_generator.__doc__)

        config = block_generator.config
        input = block_generator.input  # @ReservedAssignment
        output = block_generator.output

    else:
        assert False

    #  print block_type, module

    if source.endswith(".pyc"):
        source = source[:-1]

    if not isinstance(module, str):
        print((block_type, "has it wrong"))

    return ModelDoc(
        name=block_type,
        source=source,
        type=type,
        module=module,
        implementation=implementation,
        desc=desc,
        desc_rest=desc_rest,
        input=input,
        output=output,
        config=config,
    )


def get_all_info(library):

    blocks = []
    for block_type, generator in list(library.name2block.items()):
        blocks.append(collect_info(block_type, generator))

    # get all modules
    module_names = set([x.module for x in blocks])
    # create module -> list of blocks

    modules = {}

    for name in module_names:
        module_blocks = dict(
            [(block.name, block) for block in list([block for block in blocks if block.module == name])]
        )

        actual = __import__(name, fromlist=["ceremony"])

        desc, desc_rest = split_docstring(actual.__doc__)

        try:
            procgraph_info = get_module_info(name)
        except Exception as e:
            print(("Warning: procgraph_info missing for %r: %s" % (name, e)))
            procgraph_info = {}

        modules[name] = ModuleDoc(
            name=name,
            blocks=module_blocks,
            desc=desc,
            desc_rest=desc_rest,
            module=actual,
            procgraph_info=procgraph_info,
        )

    return modules


def main():
    parser = OptionParser()

    parser.add_option("--output", default="pgdoc.rst", help="RST output file.")
    parser.add_option("--translate", action="append", default=[], help="directory=url")

    (options, args) = parser.parse_args()

    translate = {}
    for couple in options.translate:
        root, reference = couple.split("=", 1)
        root = os.path.realpath(root)
        translate[root] = reference

    if not args:
        print("Give at least one module.")
        sys.exit(-1)

    given_modules = args

    for module in given_modules:
        __import__(module, fromlist=["ceremony"])

    library = default_library

    all_modules = get_all_info(library)

    # check they were not empty
    for module in given_modules:
        if not module in all_modules:
            print(
                (
                    "Warning: I found no blocks defined *directly* in %r. "
                    "No documentation will be generated." % module
                )
            )

    # only retain the ones that we have to document
    modules = {}
    for module in all_modules:
        for arg in args:
            if module.startswith(arg):
                modules[module] = all_modules[module]

    f = open(options.output, "w")

    f.write(".. |towrite| replace:: **to write** \n\n")

    # first write a summary
    for module_name in sorted(modules):
        module = modules[module_name]

        # f.write('%s\n' % module_reference(module.name))
        f.write("\nPackage ``%s``\n" % module.name)
        f.write("-" * 60 + "\n\n")

        d = getstr(module.desc, "%s description" % module.name)
        f.write(d + "\n\n")

        rows = []
        for block_name in sorted(module.blocks):
            block = module.blocks[block_name]
            desc = getstr(block.desc, "%s:%s desc" % (module_name, block_name))
            name = block_reference(block.name)
            rows.append((name, desc))
        write_rst_table(f, rows, widths=[30, 70])

    for module_name in sorted(modules):
        module = modules[module_name]

        f.write(module_anchor(module.name))
        f.write(rst_class("procgraph:module"))
        f.write("Package ``%s``\n" % module.name)
        f.write("=" * 60 + "\n\n\n")

        f.write(rst_class("procgraph:desc"))
        f.write(getstr(module.desc, module_name) + "\n\n")

        if module.desc_rest:
            f.write(rst_class("procgraph:desc_rest"))
            f.write(module.desc_rest + "\n\n")

        for block_name in sorted(module.blocks):
            block = module.blocks[block_name]

            f.write(block_anchor(block.name))
            f.write(rst_class("procgraph:block"))
            f.write("Block ``%s``\n" % block.name)
            f.write("-" * 66 + "\n")

            d = getstr(block.desc, "%s:%s" % (module_name, block_name))
            f.write(d + "\n\n")

            if block.desc_rest:
                f.write(block.desc_rest + "\n\n")

            if block.config:
                f.write(rst_class("procgraph:config"))
                f.write("Configuration\n")
                f.write("^" * 60 + "\n\n")
                for c in block.config:
                    d = getstr(c.desc, "%s:%s config %s" % (module_name, block_name, c.variable))
                    if c.has_default:
                        f.write("- ``%s`` (default: %s): %s\n\n" % (c.variable, c.default, d))
                    else:
                        f.write("- ``%s``: %s\n\n" % (c.variable, d))
                        # TODO: add desc_rest

            if block.input:
                f.write(rst_class("procgraph:input"))
                f.write("Input\n")
                f.write("^" * 60 + "\n\n")
                for i in block.input:
                    d = getstr(i.desc, "%s:%s input %s" % (module_name, block_name, i.name))

                    if i.type == FIXED:
                        f.write("- ``%s``: %s\n\n" % (i.name, d))
                    elif i.type == VARIABLE:

                        if i.max is None and i.min is None:
                            condition = ""
                        elif i.max is None:
                            condition = ": n >= %d" % (i.min)
                        elif i.min is None:
                            condition = ": n <= %d" % (i.max)
                        else:
                            condition = ": %d <= n <= %d" % (i.min, i.max)
                        f.write("- %s (variable number%s)\n\n" % (d, condition))
                    else:
                        raise Exception(i.type)

            if block.output:
                f.write(rst_class("procgraph:output"))
                f.write("Output\n")
                f.write("^" * 60 + "\n\n")
                for o in block.output:
                    d = getstr(o.desc, "%s:%s output %s" % (module_name, block_name, o.name))

                    if o.type == FIXED:
                        f.write("- ``%s``: %s\n\n" % (o.name, d))
                    elif o.type == VARIABLE:
                        f.write("- %s (variable number)\n\n" % d)
                    elif o.type == DEFINED_AT_RUNTIME:
                        f.write("- %s (signals are defined at runtime)\n\n" % d)
                    else:
                        raise Exception(o.type)

            url = get_source_ref(block.source, translate)
            f.write(rst_class("procgraph:source"))
            f.write("Implemented in %s. \n\n\n" % url)


def getstr(s, what):
    if s is None:
        print(("Missing documentation: %s" % what))
        return "|towrite|"
    else:
        return s


def get_source_ref(source, translation):
    source = os.path.realpath(source)
    for prefix, ref in list(translation.items()):
        if source.startswith(prefix):
            rest = source[len(prefix) :]
            url = ref + "/" + rest
            return "`%s <%s>`_" % (rest, url)
    return source


def rst_class(c):
    return "\n.. rst-class:: %s\n\n" % c


def module_anchor(name):
    return ".. _`module:%s`:\n\n" % name


def block_anchor(name):
    return ".. _`block:%s`:\n\n" % name


def block_reference(name):
    # return ":ref:`block:%s`" % name
    return ":ref:`%s <block:%s>`" % (name, name)


def module_reference(name):
    return ":ref:`module:%s`" % name


"""
.. list-table:: Frozen Delights!
   :widths: 15 10 30
   :header-rows: 1

   * - Treat
     - Quantity
     - Description
   * - Albatross
     - 2.99
     - On a stick!
   * - Crunchy Frog
     - 1.49
     - If we took the bones out, it wouldn't be
       crunchy, now would it?
   * - Gannet Ripple
     - 1.99
     - On a stick!
"""


def write_rst_table(f, rows, widths=[30, 70]):
    # XXX: does not work with multiline
    f.write(".. list-table::\n")
    f.write("   :widths: %s\n" % " ".join([str(w) for w in widths]))
    f.write("\n")

    for row in rows:
        for i, item in enumerate(row):
            if "\n" in item:
                print(("Warning: malformed cell %r." % item))
            if i == 0:
                f.write("   * - %s\n" % item)
            else:
                f.write("     - %s\n" % item)
    f.write("\n")



from . import cv
from . import np


def numpy_to_cv(numpy_array):
    """ Converts a HxW or HxWx3 numpy array to OpenCV 'image' """

    dtype2depth = {
        "uint8": cv.IPL_DEPTH_8U,
        "int8": cv.IPL_DEPTH_8S,
        "uint16": cv.IPL_DEPTH_16U,
        "int16": cv.IPL_DEPTH_16S,
        "int32": cv.IPL_DEPTH_32S,
        "float32": cv.IPL_DEPTH_32F,
        "float64": cv.IPL_DEPTH_64F,
    }

    if len(numpy_array.shape) == 2:
        (height, width) = numpy_array.shape
        nchannels = 1
    elif len(numpy_array.shape) == 3:
        (height, width, nchannels) = numpy_array.shape
    else:
        raise ValueError("Invalid format shape %s" % str(numpy_array.shape))

    im_cv = cv.CreateImage((width, height), dtype2depth[str(numpy_array.dtype)], nchannels)
    cv.SetData(im_cv, numpy_array.tostring(), numpy_array.dtype.itemsize * width * nchannels)
    return im_cv


def cv_to_numpy(im):
    """Converts opencv to numpy """
    depth2dtype = {
        cv.IPL_DEPTH_8U: "uint8",
        cv.IPL_DEPTH_8S: "int8",
        cv.IPL_DEPTH_16U: "uint16",
        cv.IPL_DEPTH_16S: "int16",
        cv.IPL_DEPTH_32S: "int32",
        cv.IPL_DEPTH_32F: "float32",
        cv.IPL_DEPTH_64F: "float64",
    }

    a = np.fromstring(im.tostring(), dtype=depth2dtype[im.depth], count=im.width * im.height * im.nChannels)
    a.shape = (im.height, im.width, im.nChannels)
    return a



import numpy as np
from procgraph import Block
from procgraph.core.registrar_other import simple_block

__all__ = ["TransparentImageAverage"]


class ChannelAverage:
    def __init__(self):
        self.value = None
        self.alpha = None

    def update(self, value, alpha):
        if self.value is None:
            self.value = value
            self.alpha = alpha
        else:
            alpha_sum = self.alpha + alpha
            alpha_sum[alpha_sum == 0] = 1
            self.value = (self.value * self.alpha + value * alpha) / alpha_sum
            self.alpha = self.alpha + alpha

    def get_value(self):
        return self.value

    def get_alpha_255(self):
        return self.alpha * 255.0 / self.alpha.max()


class TransparentImageAverage:
    """ Averages images, assuming that alpha means a weight factor. """

    def __init__(self):
        self.channels = [ChannelAverage() for _ in range(3)]

    def update(self, rgba):
        rgb = rgba[:, :, 0:3].astype("float32")
        alpha = rgba[:, :, 3].astype("float32")

        for i in range(3):
            self.channels[i].update(rgb[:, :, i], alpha)

    def get_rgb(self):
        rgbs = np.dstack([self.channels[i].get_value() for i in range(3)])
        rgbs = rgbs.astype("uint8")
        return rgbs

    def get_rgba(self):
        channels = [self.channels[i].get_value().astype("uint8") for i in range(3)]
        channels.append(self.channels[0].get_alpha_255().astype("uint8"))

        rgba = np.dstack(channels)

        return rgba


# XXX: move away
@simple_block
def to_rgb(rgba):
    return rgba[:, :, 0:3]


class TransparentImageAverageBlock(Block):
    Block.alias("trans_avg")
    Block.input("rgba")
    #     Block.output('rgb')
    Block.output("rgba")

    def init(self):
        self.tia = TransparentImageAverage()

    def update(self):
        self.tia.update(self.input.rgba)
        # self.output.rgb = self.tia.get_rgb()
        self.output.rgba = self.tia.get_rgba()



from procgraph import Block


class Sieve(Block):
    """ 
        This block decimates the data in time by transmitting
        only one in ``n`` updates.
    """

    Block.alias("sieve")

    Block.config("n", "Decimation level; ``n = 3`` means transmit one in three.")

    Block.input("data", "Arbitrary input signals.")
    Block.output("decimated", "Decimated signals.")

    def init(self):
        self.state.count = 0

    def update(self):
        # make something happen after we have waited enough
        if 0 == self.state.count % self.config.n:
            # Just copy the input to the output
            # XXX: using only one signal?
            for i in range(self.num_input_signals()):
                self.set_output(i, self.get_input(i), self.get_input_timestamp(i))

        self.state.count += 1



from procgraph import simple_block


@simple_block
def cov2corr(covariance, zero_diagonal=True):
    """ 
    Compute the correlation matrix from the covariance matrix.
    If zero_diagonal = True, the diagonal is set to 0 instead of 1. 

    :param zero_diagonal: Whether to set the (noninformative) diagonal to zero.
    :param covariance: A 2D numpy array.
    :return: correlation: The exctracted correlation.
    
    """
    from numpy import multiply, sqrt

    # TODO: add checks
    outer = multiply.outer

    sigma = sqrt(covariance.diagonal())
    M = outer(sigma, sigma)
    correlation = covariance / M

    if zero_diagonal:
        for i in range(covariance.shape[0]):
            correlation[i, i] = 0

    return correlation



import numpy as np
from procgraph import Block
from procgraph_mpl.plot_generic import PlotGeneric


__all__ = [
    "Bounce",
]


class Bounce(Block):

    Block.alias("bounce")

    Block.config("width", "Image dimension", default=320)
    Block.config("height", "Image dimension", default=240)
    Block.config("transparent", "If true, outputs a RGBA image instead of RGB.", default=False)
    Block.config("tight", 'Uses "tight" option for creating png (Matplotlib>=1.1).', default=False)

    Block.input("tick")
    Block.output("rgb")

    def init(self):
        self.plot_generic = PlotGeneric(
            width=self.config.width,
            height=self.config.height,
            transparent=self.config.transparent,
            tight=self.config.tight,
        )

    def update(self):
        self.output.rgb = self.plot_generic.get_rgb(self.plot)

    def plot(self, pylab):
        t = self.get_input_timestamp(0)
        t0 = t
        t1 = t + 2
        x = np.linspace(t0, t1, 1000)
        y = np.cos(x)
        pylab.plot(x, y)
        pylab.axis((t0, t1, -1.2, +1.2))



""" Various operations wrapping numpy functions. """


procgraph_info = {
    # List of python packages
    "requires": ["numpy"],
}

from . import filters
from . import gradient1d
from . import smooth1d



from procgraph import Block


class HistoryT(Block):
    """ 
    This block collects the signals samples of a signals,
    and outputs *one* signal containing a tuple  ``(t,x)``. 
    See also :ref:`block:last_n_samples` and :ref:`block:history`.
    
    If ``natural`` is true, it uses the time from the beginning of the log.
     
    """

    Block.alias("historyt")

    Block.config("interval", "Length of interval (seconds).", default=10)
    Block.config(
        "natural",
        "If true, set 0 to be timestamp of the log " "beginning. This allows to have prettier graphs",
        default=True,
    )

    Block.input("x", "Any signal.")
    Block.output("history", "Tuple ``(t,x)`` containing two arrays.")

    def init(self):
        self.state.x = []
        self.state.t = []
        self.state.first_timestamp = None

    def update(self):
        sample = self.get_input(0)
        timestamp = self.get_input_timestamp(0)

        if self.state.first_timestamp is None:
            self.state.first_timestamp = timestamp

        if self.config.natural:
            timestamp = timestamp - self.state.first_timestamp

        x = self.state.x
        t = self.state.t

        x.append(sample)
        t.append(timestamp)

        while abs(t[0] - t[-1]) > self.config.interval:
            t.pop(0)
            x.pop(0)

        self.output.history = (t, x)



from procgraph import Block


class Psychedelic(Block):
    Block.alias("psychedelic")
    Block.input("rgb", "An RGB image.")
    Block.output("processed", "The processed image.")

    def init(self):
        self.channel = 0

    def update(self):
        self.channel = (self.channel + 1) % 3

        rgb = self.input.rgb.copy()
        for i in [0, 1, 2]:
            if i != self.channel:
                rgb[:, :, i] = 0

        self.output.processed = rgb



from procgraph import Generator, Block

__all__ = ["Clock"]


class Clock(Generator):
    Block.alias("clock")
    Block.config("interval", "Delta between ticks.", default=1)
    Block.output("clock", "Clock signal.")
    Block.config("length", "Total interval", default=None)

    def init(self):
        self.state.clock = 0.0

    def update(self):
        self.set_output("clock", self.state.clock, timestamp=self.state.clock)
        self.state.clock += self.config.interval

    def next_data_status(self):
        if self.config.length is not None and self.state.clock > self.config.length:
            return (False, None)
        else:
            return (True, self.state.clock + self.config.interval)



from ..core.exceptions import SemanticError
from ..core.model_loader import model_from_string  # XXX:

from .utils import PGTestCase

good_examples2 = [
    """ |constant value=2| -> [0] res """,
    """ # Multiple inputs
|constant value=12| -> a
|constant value=13| -> b
a, b -> |+| -> c  """,
    """ # Multiple inputs / anonymous
|constant value=12| -> a -> |gain k=3| -> c
|constant value=13| -> |gain k=-1| -> d
 c, d -> |+| -> result         """,
    """ # Referring to outputs using numbers  
|constant value=12| -> [0]a  """,
    """ # Referring to input/outputs using numbers  
|constant value=12| -> [0]a[0] -> |gain k=1|  """,
    """ # A block by itself should be fine  
|generic_in0_out0|  
|generic_in0_out2|  """,
    """ # Trying some named connections (1)  
|generic_in0_out1| -> |generic_in1_out0|  """,
    """ # Trying some named connections (1)
# Should be the same  
|generic_in0_out1| -> x
x -> |generic_in1_out0|  """,
    """# Should be the same  as well
|c1:constant value=1| -> |g1:generic_in1_out1|
g1.0 -> |g2:generic_in1_out0|  """,
    """# Should be the same  as well
|g1:generic_in0_out1|
g1.0 -> |g2:generic_in1_out0|  """,
    #
    # """# Should be the same  as well
    # |g1:generic_in0_out1|
    # |g2:generic_in0_out1|
    # g2.0 -> g1.1  """
]

bad_examples2 = [
    """ # Bad number of outputs  
|constant value=12| -> a, b  """,
    """ # Bad output name  
|constant value=12| -> [inexistent]a  """,
    """ # Bad output number  
|constant value=12| -> [1]a  """,
    """ # inexistent input
invalid -> |gain k=2|            """,
    """ # input to something with no input
|constant value=12| -> a
a -> |constant value=12|   """,
    """ # wrong number of inputs
|constant value=12| -> a
|constant value=12| -> b
a, b -> |gain k=1|              """,
    """ # Cannot use output if terminating  
|constant value=12| -> [0]a[0]  """,
    """ # Incompatible signals (anonymous) 
|constant value=12| -> |generic in=2 out=1| -> y """,
    """ # We don't want to connect blocks with no signals  
|generic in=0 out=0| -> |generic in=0 out=2|  """,
    """ # It cannot be without input if one is needed   
|generic in=1 out=0|  """,
    """ # Double definition   
|generic_in0_out1| -> a  
|generic_in0_out1| -> a""",
    """ # Oops, definition is left dangling...   
|generic_in0_out1| -> a  
|generic_in0_out1| -> a
g1.in = 2
""",
]


class SemanticsTest(PGTestCase):
    def testBadExamples(self):
        for example in bad_examples2:
            self.check_semantic_error(example)

    def testExamples(self):
        for example in good_examples2:
            self.check_semantic_ok(example)


class ParamsTest(PGTestCase):
    def test1(self):
        """ Checking that we signal unused parameters. """

        model_desc = """
           |input name=x| -> |g1:gain| -> |output name=y|
        """
        # g2.gain is unused
        self.assertRaises(SemanticError, model_from_string, model_desc, config={"g2.k": 2})



from procgraph import simple_block
from procgraph.block_utils import check_2d_array
from procgraph_cv.conversions import cv_to_numpy, numpy_to_cv

try:
    from cv2 import cv  # @UnresolvedImport
except:
    pass


@simple_block(num_outputs=2)
def gradient(grayscale, aperture_size=3):
    """ 
        Computes the gradient of an image using a Sobel filter.
    
        :param grayscale:  A field to derive.
        :type  grayscale:  HxW array float
        
        :param aperture_size: Aperture of the Sobel filter (odd).
        :type  aperture_size: int,odd,>=1
     
        :return: gx: Gradient in the *x* direction.
        :rtype: array(HxW,float)
        
        :return: gy: Gradient in the *y* direction.
        :rtype: array(HxW,float)
    """

    check_2d_array(grayscale, "grayscale")

    im = numpy_to_cv(grayscale)
    shape = (im.width, im.height)
    sobel = cv.CreateImage(shape, cv.IPL_DEPTH_32F, 1)
    cv.Sobel(im, sobel, 1, 0, aperture_size)
    gx = cv_to_numpy(sobel).squeeze()

    sobel = cv.CreateImage(shape, cv.IPL_DEPTH_32F, 1)
    cv.Sobel(im, sobel, 0, 1, aperture_size)
    gy = cv_to_numpy(sobel).squeeze()

    return gx.astype("float32"), gy.astype("float32")


@simple_block
def smooth(grayscale, gaussian_std=5.0):
    """ 
        Smooths an image with a Gaussian filter.
        
        :param grayscale:  A field to derive.
        :type  grayscale:  HxW array float
        
        :param gaussian_std: Std-deviation of the Gaussian filter.
        :type  gaussian_std: float,>0
     
        :return: smoothed: The smoothed image.
        :rtype: array(HxW,float)
        
    """

    check_2d_array(grayscale, "grayscale")
    grayscale = grayscale.astype("float32")

    im = numpy_to_cv(grayscale)
    shape = (im.width, im.height)
    smoothed = cv.CreateImage(shape, cv.IPL_DEPTH_32F, 1)

    cv.Smooth(im, smoothed, cv.CV_GAUSSIAN, 0, 0, gaussian_std)

    result_a = cv_to_numpy(smoothed).squeeze()
    return result_a



import os


def make_sure_dir_exists(filename):
    """ Makes sure that the path to file exists, but creating directories. """
    dirname = os.path.dirname(filename)
    # dir == '' for current dir
    if dirname != "" and not os.path.exists(dirname):
        os.makedirs(dirname)


def expand(filename):
    """ Expands environment variables in the filename. """
    filename = os.path.expandvars(filename)
    filename = os.path.expanduser(filename)
    return filename



from procgraph import Block

__all__ = ["History"]


class History(Block):
    """ 
    This block collects the history of a quantity,
    and outputs two signals ``x`` and ``t``. 
    See also :ref:`block:historyt` and :ref:`block:last_n_samples`.
    """

    Block.alias("history")

    Block.config("interval", "Length of the interval to record.", default=10)

    Block.input("values", "Any signal.")

    Block.output("x", "Sequence of values.")
    Block.output("t", "Sequence of timestamps.")

    def init(self):
        self.history = HistoryInterval(self.config.interval)

    def update(self):
        sample = self.get_input(0)
        timestamp = self.get_input_timestamp(0)

        self.history.push(timestamp, sample)
        ts, xs = self.history.get_ts_xs()
        self.output.x = xs
        self.output.t = ts


class HistoryInterval(object):
    def __init__(self, interval):
        self.interval = interval
        self.x = []
        self.t = []

    def push(self, timestamp, value):

        self.x.append(value)
        self.t.append(timestamp)

        while abs(self.t[0] - self.t[-1]) > self.interval:
            self.t.pop(0)
            self.x.pop(0)

    def get_ts_xs(self):
        return list(self.t), list(self.x)



from .strings import *
from .friendly_paths import *
from .calling_ext_program import *
from .script_utils import *
from .inawhile import *
from .safe_pickle import *
from .levenshtein_match import *



from . import np  # @PydevCodeAnalysisIgnore
from procgraph import Block


class Expectation(Block):
    """ Computes the sample expectation of a signal. """

    Block.alias("expectation")

    Block.input("x", "Any numpy array.")
    Block.output("Ex", "Expectation of input.")

    def init(self):
        self.state.num_samples = 0

    def update(self):
        N = self.state.num_samples

        if N == 0:
            self.state.Ex = self.input.x.copy()
        else:
            self.state.Ex = (self.state.Ex * N + self.input.x) / float(N + 1)

        self.state.num_samples += 1
        self.output.Ex = self.state.Ex


class ExpectationNorm(Block):
    Block.alias("expectation_norm")

    Block.input("x", "Any numpy array.")
    Block.output("Ex", "Expectation of input.")

    def init(self):
        self.value = None
        self.mass = None

    def update(self):
        x = self.input.x
        if self.value is None:
            self.value = x
            self.weight = np.zeros(x.shape)
        else:
            if False:
                w = np.abs(x - self.last_x)
            else:
                w = np.abs(x - self.last_x).sum()
            self.weight += w
            self.value += w * x

            self.weight[self.weight == 0] = np.inf
            res = self.value / self.weight
            self.output.Ex = res
        self.last_x = x.copy()
